#!/usr/bin/python
# cmd_line_json.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (cmd_line_json.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: cmd_line_json
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import os
import sys

#------------------------------------------------------------------------------

from bitdust.lib import jsontemplate
from bitdust.lib import strng
from bitdust.lib import jsn

from bitdust.interface import cmd_line_json_templates as templ

#------------------------------------------------------------------------------


def parser():
    """
    Create an ``optparse.OptionParser`` object to read command line arguments.
    """
    from optparse import OptionParser, OptionGroup
    from bitdust.main.help import usage_text
    parser = OptionParser(usage=usage_text())
    group = OptionGroup(parser, 'Logs')
    group.add_option(
        '-d',
        '--debug',
        dest='debug',
        type='int',
        help='set debug level',
    )
    group.add_option(
        '-q',
        '--quite',
        dest='quite',
        action='store_true',
        help='quite mode, do not print any messages to stdout',
    )
    group.add_option(
        '-v',
        '--verbose',
        dest='verbose',
        action='store_true',
        help='verbose mode, print more messages',
    )
    group.add_option(
        '-n',
        '--no-logs',
        dest='no_logs',
        action='store_true',
        help='do not use logs',
    )
    group.add_option(
        '-o',
        '--output',
        dest='output',
        type='string',
        help='print log messages to the file',
    )
    group.add_option(
        '--twisted',
        dest='twisted',
        action='store_true',
        help='show twisted log messages too',
    )
    parser.add_option_group(group)
    return parser


def override_options(opts, args):
    """
    The program can replace some user options by values passed via command
    line.

    This method return a dictionary where is stored a key-value pairs
    for new options.
    """
    overDict = {}
    if opts.debug or str(opts.debug) == '0':
        overDict['logs.debug-level'] = str(opts.debug)
    return overDict


#------------------------------------------------------------------------------


def print_copyright():
    """
    Prints the copyright string.
    """
    print_text('Copyright BitDust, 2014. All rights reserved.')


def print_text(msg, nl='\n'):
    """
    Send some output to the console.
    """
    sys.stdout.write(strng.to_text(msg) + nl)
    sys.stdout.flush()


def print_exception():
    """
    This is second most common method in the project.

    Print detailed info about last exception to the logs.
    """
    import traceback
    _, value, trbk = sys.exc_info()
    try:
        excArgs = value.__dict__['args']
    except KeyError:
        excArgs = ''
    excTb = traceback.format_tb(trbk)
    s = 'Exception: <' + str(value) + '>\n'
    if excArgs:
        s += '  args:' + excArgs + '\n'
    for l in excTb:
        s += l + '\n'
    sys.stdout.write(s)
    sys.stdout.flush()


def print_and_stop(result):
    """
    Print text to console and stop the reactor.
    """
    from twisted.internet import reactor  # @UnresolvedImport
    sys.stdout.write(jsn.dumps(result, indent=2, keys_to_text=True, values_to_text=True, ensure_ascii=True) + '\n')
    reactor.stop()  # @UndefinedVariable


def print_template(result, template):
    """
    Use json template to format the text and print to STDOUT.
    """
    try:
        template.undefined_str = '?'
        raw = template.expand(result)
    except Exception as e:
        print('failed to render output data: %r' % e)
        return
    sys.stdout.write(raw)
    sys.stdout.flush()


def print_template_and_stop(result, template):
    """
    Print text with json template formatting and stop the reactor.
    """
    from twisted.internet import reactor  # @UnresolvedImport
    print_template(result, template)
    reactor.stop()  # @UndefinedVariable


def fail_and_stop(err):
    """
    Send error message to STDOUT and stop the reactor.
    """
    from twisted.internet import reactor  # @UnresolvedImport
    try:
        print_text(err.getErrorMessage())
    except:
        print(err)
    reactor.stop()  # @UndefinedVariable


#------------------------------------------------------------------------------


def call_websocket_method(method, **kwargs):
    """
    Method allows to communicate with an existing BitDust process via WebSocket API interface.
    """
    from twisted.internet.defer import Deferred  # @UnresolvedImport
    from twisted.internet import reactor  # @UnresolvedImport
    from bitdust.lib import websock
    from bitdust.system import deploy
    ret = Deferred()
    timeout = kwargs.pop('websocket_timeout', None)
    if timeout:
        ret.addTimeout(timeout=(timeout + 1), clock=reactor)

    def _on_result(resp):
        if _Debug:
            print('call_websocket_method._on_result', method, kwargs, resp)
        try:
            websock.stop()
        except Exception as exc:
            if _Debug:
                print('call_websocket_method._on_result', exc)
        if method in [
            'process_stop',
        ]:
            if not ret.called:
                ret.callback({'status': 'OK'})
            return None
        if not isinstance(resp, dict):
            if not ret.called:
                ret.errback(resp)
            return None
        if 'payload' not in resp:
            if not ret.called:
                ret.errback(Exception('received an empty response'))
            return None
        if not isinstance(resp['payload'], dict):
            if not ret.called:
                ret.errback(Exception('unexpected response received: %r' % resp['payload']))
            return None
        if 'errors' in resp['payload']:
            if not ret.called:
                ret.errback(Exception(', '.join(resp['payload']['errors'])))
            return None
        try:
            payload_response = resp['payload']['response']
        except Exception as exc:
            ret.errback(exc)
            return None
        ret.callback(payload_response)
        return resp

    def _on_open(ws_inst):
        if _Debug:
            print('call_websocket_method._on_open', method, kwargs, ws_inst)
        websock.ws_call(
            json_data={
                'command': 'api_call',
                'method': method,
                'kwargs': kwargs,
            },
            cb=_on_result,
            timeout=timeout,
        )

    def _on_error(err):
        if _Debug:
            print('call_websocket_method._on_error', method, kwargs, err)
        try:
            websock.stop()
        except Exception as exc:
            if _Debug:
                print('call_websocket_method._on_error', exc)
        if method in [
            'process_stop',
        ]:
            return None
        if not ret.called:
            ret.errback(err)
        return None

    websock.start(
        callbacks={
            'on_open': _on_open,
            'on_error': _on_error,
        },
        api_secret_filepath=os.path.join(deploy.current_base_dir(), 'apisecret'),
    )
    return ret


def call_websocket_method_and_stop(method, **kwargs):
    from twisted.internet import reactor  # @UnresolvedImport
    d = call_websocket_method(method, **kwargs)
    d.addCallback(print_and_stop)
    d.addErrback(fail_and_stop)
    reactor.run()  # @UndefinedVariable
    return 0


def call_websocket_method_template_and_stop(method, template, **kwargs):
    from twisted.internet import reactor  # @UnresolvedImport
    d = call_websocket_method(method, **kwargs)
    d.addCallback(print_template_and_stop, template)
    d.addErrback(fail_and_stop)
    reactor.run()  # @UndefinedVariable
    return 0


def call_websocket_method_transform_template_and_stop(method, template, transform, **kwargs):
    from twisted.internet import reactor  # @UnresolvedImport
    d = call_websocket_method(method, **kwargs)
    d.addCallback(lambda result: print_template_and_stop(transform(result), template))
    d.addErrback(fail_and_stop)
    reactor.run()  # @UndefinedVariable
    return 0


def call_api_method_template_and_stop(method, template, **kwargs):
    from twisted.internet import reactor  # @UnresolvedImport
    from bitdust.main import settings
    from bitdust.interface import api

    def _run():
        settings.init()
        m = getattr(api, method)
        d = m(**kwargs)
        d.addCallback(print_template_and_stop, template)
        d.addErrback(fail_and_stop)

    reactor.callWhenRunning(_run)  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable
    return 0


#------------------------------------------------------------------------------


def kill():
    """
    Kill all running BitDust processes (except current).
    """
    import time
    from bitdust.system import bpio
    total_count = 0
    found = False
    while True:
        appList = bpio.lookup_main_process()
        if len(appList) > 0:
            found = True
        for pid in appList:
            print_text('trying to stop pid %d' % pid)
            bpio.kill_process(pid)
        if len(appList) == 0:
            if found:
                print_text('BitDust stopped\n')
            else:
                print_text('BitDust was not started\n')
            return 0
        total_count += 1
        if total_count > 10:
            print_text('some BitDust process found, but can not stop it\n')
            return 1
        time.sleep(1)


def wait_then_kill(x):
    """
    For correct shutdown of the program need to send a URL request to the HTTP
    server:: http://localhost:<random port>/?action=exit.

    After receiving such request the program will call
    ``p2p.init_shutdown.shutdown()`` method and stops. But if the main
    process was blocked it needs to be killed with system "kill"
    procedure. This method will wait for 10 seconds and then call method
    ``kill()``.
    """
    import time
    from twisted.internet import reactor  # @UnresolvedImport
    from bitdust.system import bpio
    total_count = 0
    while True:
        appList = bpio.lookup_main_process()
        if len(appList) == 0:
            print_text('finished successfully')
            reactor.stop()  # @UndefinedVariable
            return 0
        total_count += 1
        if total_count > 10:
            print_text('not responding, going to kill the running process ...')
            ret = kill()
            reactor.stop()  # @UndefinedVariable
            return ret
        time.sleep(1)


#------------------------------------------------------------------------------


def run_now(opts, args):
    from bitdust.system import bpio
    from bitdust.logs import lg
    lg.life_begins()
    if opts.no_logs:
        lg.disable_logs()
    overDict = override_options(opts, args)
    from bitdust.main.bpmain import run
    ui = ''
    if len(args) > 0 and args[0].lower() in [
        'restart',
    ]:
        ui = 'show'
    ret = run(ui, opts, args, overDict)
    bpio.shutdown()
    return ret


#------------------------------------------------------------------------------


def cmd_deploy(opts, args, overDict):
    from bitdust.system import deploy
    return deploy.run(args)


#------------------------------------------------------------------------------


def cmd_reconnect(opts, args, overDict):
    tpl = jsontemplate.Template(templ.TPL_RAW)
    return call_websocket_method_template_and_stop('network_reconnect', tpl)


#------------------------------------------------------------------------------


def cmd_network(opts, args, overDict, running):
    if len(args) >= 3 and args[1] in ['create', 'setup']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        if not running:
            return call_api_method_template_and_stop('network_create', tpl, url=args[2])
        return call_websocket_method_template_and_stop('network_create', tpl, url=args[2])

    if len(args) >= 3 and args[1] in ['select', 'chose', 'set', 'switch']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        if not running:
            return call_api_method_template_and_stop('network_select', tpl, name=args[2])
        return call_websocket_method_template_and_stop('network_select', tpl, name=args[2])

    return 2


#------------------------------------------------------------------------------


def cmd_identity(opts, args, overDict, running, executablePath):
    from bitdust.main import initializer
    from bitdust.main import shutdowner
    from bitdust.main import settings
    from bitdust.userid import my_id
    initializer.init_settings()
    initializer.init_automats()
    my_id.init()

    def _do_cmd():
        if args[0] == 'idurl' and len(args) <= 1:
            if my_id.isLocalIdentityReady():
                print_text(my_id.getIDURL())
            else:
                print_text('local identity is not valid or not exist')
            return 0

        if args[0] in ['globid', 'globalid', 'gid', 'glid'] or (args[0] == 'id' and len(args) <= 1):
            if my_id.isLocalIdentityReady():
                print_text(my_id.getGlobalID())
            else:
                print_text('local identity is not valid or not exist')
            return 0

        if len(args) == 1 or args[1].lower() in ['info', '?', 'show', 'print']:
            if my_id.isLocalIdentityReady():
                print_text(my_id.getLocalIdentity().serialize(as_text=True))
            else:
                print_text('local identity is not valid or not exist')
            return 0

        from twisted.internet import reactor  # @UnresolvedImport

        if args[1] in ['server', 'srv'] and args[0]:

            def _run_stand_alone_id_server():
                from bitdust.logs import lg
                from bitdust.userid import id_server
                lg.open_log_file(os.path.join(settings.LogsDir(), 'idserver.log'))
                lg.set_debug_level(settings.getDebugLevel())
                reactor.addSystemEventTrigger('before', 'shutdown', id_server.A().automat, 'shutdown')  # @UndefinedVariable
                reactor.callWhenRunning(  # @UndefinedVariable
                    id_server.A,
                    'init',
                    (settings.getIdServerWebPort(), settings.getIdServerTCPPort()),
                )
                reactor.callLater(0, id_server.A, 'start')  # @UndefinedVariable
                reactor.run()  # @UndefinedVariable

            if len(args) <= 2:
                if not running:
                    _run_stand_alone_id_server()
                    return 0
                tpl = jsontemplate.Template(templ.TPL_SERVICE_INFO)
                return call_websocket_method_template_and_stop('service_info', tpl, service_name='service_identity_server')
            if args[2] == 'stop':
                if not running:
                    print_text('BitDust is not running at the moment\n')
                    return 0
                tpl = jsontemplate.Template(templ.TPL_RAW)
                return call_websocket_method_template_and_stop('service_stop', tpl, service_name='service_identity_server')
            if args[2] == 'start':
                if not running:
                    _run_stand_alone_id_server()
                    return 0
                tpl = jsontemplate.Template(templ.TPL_RAW)
                return call_websocket_method_template_and_stop('service_start', tpl, service_name='service_identity_server')
            return 2

        def _register():
            if len(args) <= 2:
                return 2
            pksize = settings.getPrivateKeySize()
            if len(args) > 3:
                try:
                    pksize = int(args[3])
                except:
                    print_text('incorrect private key size\n')
                    return 0
            from bitdust.lib import misc
            if not misc.ValidUserName(args[2]):
                print_text('invalid user name')
                return 0
            # automat.LifeBegins(lg.when_life_begins())
            # automat.OpenLogFile(settings.AutomatsLog())
            initializer.A('run-cmd-line-register', {'username': args[2], 'pksize': pksize})
            reactor.run()  # @UndefinedVariable
            shutdowner.shutdown_automats()
            my_id.loadLocalIdentity()
            if my_id.isLocalIdentityReady():
                print_text('\n' + my_id.getLocalIdentity().serialize(as_text=True))
            else:
                print_text('identity creation failed, please try again later')
            return 0

        def _recover():
            from bitdust.system import bpio
            from bitdust.lib import nameurl
            if len(args) < 3:
                return 2
            src = bpio.ReadTextFile(args[2])
            if len(src) > 1024*10:
                print_text('file is too big for private key')
                return 0
            try:
                lines = src.split('\n')
                idurl = lines[0]
                txt = '\n'.join(lines[1:])
                if idurl != nameurl.FilenameUrl(nameurl.UrlFilename(idurl)):
                    idurl = ''
                    txt = src
            except:
                idurl = ''
                txt = src
            if not idurl and len(args) > 3:
                idurl = args[3]
            if not idurl:
                print_text('BitDust need to know your IDURL to recover your account\n')
                return 2
            # from bitdust.automats import automat
            # from bitdust.main import initializer
            # from bitdust.main import config
            # from bitdust.logs import lg
            initializer.init_automats()
            #             automat.LifeBegins(lg.when_life_begins())
            #             automat.SetGlobalLogEvents(config.conf().getBool('logs/automat-events-enabled'))
            #             automat.SetGlobalLogTransitions(config.conf().getBool('logs/automat-transitions-enabled'))
            #             automat.SetExceptionsHandler(lg.exc)
            #             automat.SetLogOutputHandler(lambda debug_level, message: lg.out(debug_level, message, log_name='state'))
            #             # automat.OpenLogFile(settings.AutomatsLog())
            initializer.A('run-cmd-line-recover', {'idurl': idurl, 'keysrc': txt})
            reactor.run()  # @UndefinedVariable
            shutdowner.shutdown_automats()
            my_id.loadLocalIdentity()
            if my_id.isLocalIdentityReady():
                print_text('\n' + my_id.getLocalIdentity().serialize(as_text=True))
            else:
                print_text('identity recovery FAILED')
            return 0

        if args[1].lower() in ['create', 'new', 'register', 'generate']:
            if my_id.isLocalIdentityReady():
                print_text('local identity [%s] already exist\n' % my_id.getIDName())
                return 1
            if running:
                print_text('BitDust is running at the moment, need to stop the software first\n')
                return 0
            return _register()

        if len(args) >= 2 and args[1].lower() in ['bk', 'backup', 'save']:
            from bitdust.interface import api
            key_id = 'master'
            key_json = api.key_get(key_id=key_id, include_private=True)
            if key_json['status'] != 'OK':
                print_text('\n'.join(key_json['errors']))
                return 1
            TextToSave = key_json['result']['creator'] + u'\n' + key_json['result']['private']
            if args[1] in ['bk', 'backup', 'save']:
                from bitdust.system import bpio
                curpath = os.getcwd()
                os.chdir(executablePath)
                if len(args) >= 3:
                    filenameto = bpio.portablePath(args[2])
                else:
                    key_file_name = key_json['result']['key_id'] + '.key'
                    filenameto = bpio.portablePath(os.path.join(os.path.expanduser('~'), key_file_name))
                    # filenameto = bpio.portablePath(os.path.join(executablePath, key_json['result']['key_id'] + '.key'))
                os.chdir(curpath)
                if not bpio.WriteTextFile(filenameto, TextToSave):
                    del TextToSave
                    print_text('error writing to %s\n' % filenameto)
                    return 1
                print_text('IDURL "%s" and "master" key was stored in "%s"' % (key_json['result']['creator'], filenameto))
                return 0
            return 2

        if args[1].lower() in ['restore', 'recover', 'read', 'load']:
            if running:
                print_text('BitDust is running at the moment, need to stop the software first\n')
                return 0
            return _recover()

        if args[1].lower() in ['delete', 'remove', 'erase', 'del', 'rm', 'kill']:
            if running:
                print_text('BitDust is running at the moment, need to stop the software first\n')
                return 0
            oldname = my_id.getIDName()
            if not oldname:
                print_text('local identity is not exist\n')
                return 0
            my_id.forgetLocalIdentity()
            my_id.eraseLocalIdentity()
            print_text('local identity [%s] is no longer exist\n' % oldname)
            return 0

        return 2

    ret = _do_cmd()
    shutdowner.shutdown_settings()
    return ret


#------------------------------------------------------------------------------


def cmd_key(opts, args, overDict, running, executablePath):
    if not running:
        print_text('BitDust is not running at the moment\n')
        return 0

    if len(args) == 1 or (len(args) == 2 and args[1] in ['list', 'ls']):
        tpl = jsontemplate.Template(templ.TPL_KEYS_LIST)
        return call_websocket_method_template_and_stop('keys_list', tpl, include_private=False)

    if len(args) >= 3 and args[1] in ['create', 'new', 'gen', 'generate', 'make']:
        key_id = args[2]
        key_sz = 4096
        if len(args) > 3:
            key_sz = int(args[3])
        tpl = jsontemplate.Template(templ.TPL_KEY_CREATE)
        return call_websocket_method_template_and_stop('key_create', tpl, key_alias=key_id, key_size=key_sz)

    if len(args) >= 2 and args[1] in ['copy', 'cp', 'bk', 'backup', 'save']:
        from twisted.internet import reactor  # @UnresolvedImport

        def _on_key(key_json):
            TextToSave = key_json['result']['creator'] + u'\n' + key_json['result']['private']
            if len(args) >= 4 and args[1] in ['bk', 'backup', 'save']:
                from bitdust.system import bpio
                curpath = os.getcwd()
                os.chdir(executablePath)
                filenameto = bpio.portablePath(args[3])
                os.chdir(curpath)
                if not bpio.WriteTextFile(filenameto, TextToSave):
                    del TextToSave
                    print_text('error writing to %s\n' % filenameto)
                    reactor.stop()  # @UndefinedVariable
                    return 1
                print_text('private key "%s" was stored in "%s"' % (key_json['result']['key_id'], filenameto))
            else:
                from bitdust.lib import misc
                misc.setClipboardText(TextToSave)
                print_text('key "%s" was sent to clipboard, you can use Ctrl+V to paste your private key where you want' % key_json['result']['key_id'])
            del TextToSave
            if key_json['result']['alias'] == 'master':
                print_text('WARNING! keep your "master" key in safe place, do not publish it!\n')
            reactor.stop()  # @UndefinedVariable
            return

        key_id = 'master' if len(args) < 3 else args[2]
        d = call_websocket_method('key_get', key_id=key_id, include_private=True)
        d.addCallback(_on_key)
        d.addErrback(fail_and_stop)
        reactor.run()  # @UndefinedVariable
        return 0

    if len(args) >= 2 and args[1] in ['print', 'get', 'show']:
        tpl = jsontemplate.Template(templ.TPL_KEY_GET)
        key_id = 'master' if len(args) < 3 else args[2]
        return call_websocket_method_template_and_stop('key_get', tpl, key_id=key_id, include_private=True)

    if len(args) >= 3 and args[1] in ['delete', 'erase', 'remove', 'clear', 'del', 'rm', 'kill']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('key_erase', tpl, key_id=args[2])

    if len(args) >= 4 and args[1] in ['share', 'send', 'transfer', 'access']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('key_share', tpl, key_id=args[2], idurl=args[3])

    return 2


#------------------------------------------------------------------------------


def cmd_api(opts, args, overDict, executablePath):
    if len(args) < 2:
        try:
            import inspect
            from bitdust.interface import api
        except:
            print_text('failed to import "interface.api" module')
            return 1
        for item in dir(api):
            if item.startswith('_'):
                continue
            if item in [
                'ERROR',
                'OK',
                'RESULT',
                'driver',
                'lg',
                'absolute_import',
                'Failure',
                'Deferred',
                'os',
                'time',
                'on_api_result_prepared',
                'succeed',
                'sys',
                'strng',
                'map',
                'jsn',
                'json',
                'gc',
                'config',
            ]:
                continue
            method = getattr(api, item, None)
            if not method:
                continue
            try:
                params = inspect.getargspec(method)
            except:
                print_text('    %s()' % item)
                continue
            doc_line = method.__doc__
            if not doc_line:
                doc_line = ''
            else:
                doc_line = doc_line.strip().split('\n')[0]
            print_text('\n    %s(%s)' % (
                item,
                ', '.join(params.args),
            ))
            print_text('        %s' % doc_line)
        return 0

    def _clean_value(v):
        if isinstance(v, str) and v in ['true', 'True']:
            return True
        return v

    try:
        pairs = [i.split('=') for i in args[2:]]
        kwargs = {k: _clean_value(v) for k, v in pairs}
    except Exception as e:
        print_text('failed reading input arguments: %s\n' % e)
        return 1
    return call_websocket_method_and_stop(args[1], **kwargs)


#------------------------------------------------------------------------------


def cmd_file(opts, args, overDict, executablePath):
    if args[0] in [
        'dir',
        'folder',
    ]:
        if len(args) > 2 and args[1] in ['create', 'make', 'cr', 'mk', 'add', 'bind', 'map']:
            tpl = jsontemplate.Template(templ.TPL_RAW)
            return call_websocket_method_template_and_stop('file_create', tpl, remote_path=args[2], as_folder=True)
        return 2

    if len(args) < 2 or args[1] in ['list', 'ls']:
        remote_path = args[2] if len(args) > 2 else None
        tpl = jsontemplate.Template(templ.TPL_BACKUPS_LIST)
        return call_websocket_method_template_and_stop('files_list', tpl, remote_path=remote_path)

#     if len(args) == 2 and args[1] in ['idlist', 'ids']:
#         tpl = jsontemplate.Template(templ.TPL_BACKUPS_LIST_IDS)
#         return call_websocket_method_template_and_stop('backups_id_list', tpl)

    if len(args) == 2 and args[1] in ['update', 'upd', 'refresh', 'sync']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('files_sync', tpl)

    if len(args) >= 2 and args[1] in ['running', 'progress', 'status', 'prog']:
        if len(args) >= 3:
            if args[2] in ['download', 'down']:
                tpl = jsontemplate.Template(templ.TPL_BACKUPS_TASKS_LIST)
                return call_websocket_method_template_and_stop('files_downloads', tpl)
            elif args[2] in ['upload', 'up']:
                tpl = jsontemplate.Template(templ.TPL_BACKUPS_RUNNING_LIST)
                return call_websocket_method_template_and_stop('files_uploads', tpl)
            return 2
        tpl = jsontemplate.Template(templ.TPL_BACKUPS_RUNNING_LIST)
        return call_websocket_method_template_and_stop('files_uploads', tpl)

    if len(args) > 2 and args[1] in ['create', 'make', 'cr', 'mk', 'add', 'bind', 'map']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('file_create', tpl, remote_path=args[2], as_folder=False)

    if len(args) > 3 and args[1] in ['upload', 'up', 'store', 'start', 'send', 'write']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        if not os.path.exists(os.path.abspath(args[2])):
            print_text('path %s not exist\n' % args[2])
            return 1
        return call_websocket_method_template_and_stop('file_upload_start', tpl, local_path=args[2], remote_path=args[3], wait_result=False)

    if len(args) > 2 and args[1] in ['download', 'down', 'load', 'request', 'read', 'restore']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        if len(args) > 3:
            local_path = args[3]
        else:
            local_path = os.path.join(os.getcwd(), os.path.basename(args[2]))
        return call_websocket_method_template_and_stop('file_download_start', tpl, remote_path=args[2], destination_path=local_path)

    if len(args) > 2 and args[1] in ['delete', 'del', 'rm', 'remove', 'erase']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('file_delete', tpl, remote_path=args[2])

    if len(args) >= 3 and args[1] in ['cancel', 'abort']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        if len(args) > 3:
            if args[2] in ['download', 'down']:
                return call_websocket_method_template_and_stop('file_download_stop', tpl, remote_path=args[3])
            elif args[2] in ['upload', 'up']:
                return call_websocket_method_template_and_stop('file_upload_stop', tpl, remote_path=args[3])
            return 2
        return call_websocket_method_template_and_stop('file_upload_stop', tpl, remote_path=args[2])

#     if len(args) == 2:
#         tpl = jsontemplate.Template(templ.TPL_RAW)
#         if packetid.Valid(args[1]):
#             return call_websocket_method_template_and_stop('backup_start_id', tpl, args[1])
#         if not os.path.exists(os.path.abspath(args[1])):
#             print_text('path %s not exist\n' % args[1])
#             return 1
#         return call_websocket_method_template_and_stop('backup_start_path', tpl, args[1])

    return 2


# def cmd_restore(opts, args, overDict, executablePath):
#     if len(args) < 2 or args[1] in ['list', 'ls']:
#         tpl = jsontemplate.Template(templ.TPL_BACKUPS_LIST_IDS)
#         return call_websocket_method_template_and_stop('backups_id_list', tpl)
#
#     if len(args) >= 2 and args[1] in ['running', 'progress', 'status']:
#         tpl = jsontemplate.Template(templ.TPL_RESTORES_RUNNING_LIST)
#         return call_websocket_method_template_and_stop('restores_running', tpl)
#
#     tpl = jsontemplate.Template(templ.TPL_RAW)
#     if len(args) > 2 and args[1] in ['cancel', 'abort']:
#         return call_websocket_method_template_and_stop('restore_abort', tpl, args[2])
#
#     if len(args) > 3 and args[1] in ['start', 'go', 'run', ]:
#         return call_websocket_method_template_and_stop('restore_single', tpl, args[2], args[3])
#
#     if len(args) == 2:
#         return call_websocket_method_template_and_stop('restore_single', tpl, args[1])
#
#     if len(args) == 3:
#         return call_websocket_method_template_and_stop('restore_single', tpl, args[1], args[2])
#
#     return 2

#------------------------------------------------------------------------------


def option_name_to_path(name, default=''):
    path = default
    if name in ['donated', 'shared', 'given']:
        path = 'services/supplier/donated-space'
    elif name in ['needed']:
        path = 'services/customer/needed-space'
    elif name in ['suppliers']:
        path = 'services/customer/suppliers-number'
    elif name in ['debug']:
        path = 'logs/debug-level'
    elif name in ['block-size']:
        path = 'services/backups/block-size'
    elif name in ['block-size-max', 'max-block-size']:
        path = 'services/backups/max-block-size'
    elif name in ['max-backups', 'max-copies', 'copies']:
        path = 'services/backups/max-copies'
    elif name in ['local-backups', 'local-data', 'keep-local-data']:
        path = 'services/backups/keep-local-copies-enabled'
    elif name in ['tcp']:
        path = 'services/tcp-transport/enabled'
    elif name in ['tcp-port']:
        path = 'services/tcp-connections/tcp-port'
    elif name in ['udp']:
        path = 'services/udp-transport/enabled'
    elif name in ['udp-port']:
        path = 'services/udp-datagrams/udp-port'
    elif name in ['proxy']:
        path = 'services/proxy-transport/enabled'
    elif name in ['http']:
        path = 'services/http-transport/enabled'
    elif name in ['http-port']:
        path = 'services/http-connections/http-port'
    elif name in ['dht-port']:
        path = 'services/entangled-dht/udp-port'
    elif name in ['limit-send']:
        path = 'services/network/send-limit'
    elif name in ['limit-receive']:
        path = 'services/network/receive-limit'
    elif name in ['weblog']:
        path = 'logs/stream-enable'
    elif name in ['weblog-port']:
        path = 'logs/stream-port'
    return path


def cmd_set(opts, args, overDict):
    from bitdust.main import initializer
    from bitdust.main import shutdowner
    from bitdust.interface import api
    initializer.init_settings()
    name = args[1].lower()
    if name in ['list', 'ls', 'all', 'show', 'print']:
        # sort = True if (len(args) > 2 and args[2] in ['sort', 'sorted', ]) else False
        sort = True
        result = api.configs_list(sort=sort)
        for i in range(len(result['result'])):
            val = result['result'][i]['value']
            if strng.is_string(val) and len(val) > 60:
                result['result'][i]['value'] = val[:60].replace('\n', '') + '...'
        tpl = jsontemplate.Template(templ.TPL_OPTIONS_LIST_KEY_TYPE_VALUE)
        print_template(result, tpl)
        shutdowner.shutdown_settings()
        return 0
    path = '' if len(args) < 2 else args[1]
    path = option_name_to_path(name, path)
    if path != '':
        if len(args) > 2:
            value = ' '.join(args[2:])
            result = api.config_set(path, strng.text_type(value))
        else:
            result = api.config_get(path)
        tpl = jsontemplate.Template(templ.TPL_OPTION_MODIFIED)
        print_template(result, tpl)
        shutdowner.shutdown_settings()
        return 0

    shutdowner.shutdown_settings()
    return 2


def cmd_set_request(opts, args, overDict):
    print_text('connecting to already started BitDust process ...')
    name = args[1].lower()
    if name in ['list', 'ls', 'all', 'show', 'print']:
        # sort = True if (len(args) > 2 and args[2] in ['sort', 'sorted', ]) else False
        sort = True
        tpl = jsontemplate.Template(templ.TPL_OPTIONS_LIST_KEY_TYPE_VALUE)

        def _limit_length(result):
            for i in range(len(result['result'])):
                val = result['result'][i]['value']
                if strng.is_string(val) and len(val) > 60:
                    result['result'][i]['value'] = val[:60].replace('\n', '') + '...'
            return result

        return call_websocket_method_transform_template_and_stop('configs_list', tpl, _limit_length, sort=sort)
    path = '' if len(args) < 2 else args[1]
    path = option_name_to_path(name, path)
    if len(args) == 2:
        tpl = jsontemplate.Template(templ.TPL_OPTION_SINGLE)
        return call_websocket_method_template_and_stop('config_get', tpl, key=path)
    value = ' '.join(args[2:])
    tpl = jsontemplate.Template(templ.TPL_OPTION_MODIFIED)
    return call_websocket_method_template_and_stop('config_set', tpl, key=path, value=value)


#------------------------------------------------------------------------------


def cmd_suppliers(opts, args, overDict):
    if len(args) < 2 or args[1] in ['list', 'ls']:
        tpl = jsontemplate.Template(templ.TPL_SUPPLIERS)
        return call_websocket_method_template_and_stop('suppliers_list', tpl)

    elif args[1] in ['ping', 'test', 'call', 'cl']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('suppliers_ping', tpl)

    elif args[1] in ['fire', 'replace', 'rep', 'rp'] and len(args) >= 3:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('supplier_change', tpl, position=args[2])

    elif args[1] in ['hire', 'change', 'ch'] and len(args) >= 4:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('supplier_change', tpl, position=args[2], new_supplier_id=args[3])

    return 2


#------------------------------------------------------------------------------


def cmd_customers(opts, args, overDict):
    if len(args) < 2 or args[1] in ['list', 'ls']:
        tpl = jsontemplate.Template(templ.TPL_CUSTOMERS)
        return call_websocket_method_template_and_stop('customers_list', tpl)

    elif args[1] in ['ping', 'test', 'call', 'cl']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('customers_ping', tpl)

    elif args[1] in ['reject', 'refuse', 'remove', 'delete', 'rm', 'free', 'del'] and len(args) >= 3:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('customer_reject', tpl, customer_id=args[2])

    return 2


#------------------------------------------------------------------------------


def cmd_storage(opts, args, overDict):
    if len(args) < 2:
        from twisted.internet import reactor  # @UnresolvedImport

        def _got_local(result3, result2, result1):
            result = {
                'status': 'OK',
                'execution': float(result1['execution']) + float(result2['execution']) + float(result3['execution']),
                'result': [{
                    'donated': result1['result'],
                    'consumed': result2['result'],
                    'local': result3['result'],
                }],
            }
            print_template(result, jsontemplate.Template(templ.TPL_STORAGE))
            reactor.stop()  # @UndefinedVariable

        def _got_needed(result2, result1):
            d2 = call_websocket_method('space_local')
            d2.addCallback(_got_local, result2, result1)
            d2.addErrback(fail_and_stop)

        def _got_donated(result1):
            d2 = call_websocket_method('space_consumed')
            d2.addCallback(_got_needed, result1)
            d2.addErrback(fail_and_stop)

        d1 = call_websocket_method('space_donated')
        d1.addCallback(_got_donated)
        d1.addErrback(fail_and_stop)
        reactor.run()  # @UndefinedVariable
        return 0
    return 2


#------------------------------------------------------------------------------


def cmd_automats(opts, args, overDict):
    if len(args) < 2 or args[1] == 'list':
        tpl = jsontemplate.Template(templ.TPL_AUTOMATS)
        return call_websocket_method_template_and_stop('automats_list', tpl)
    return 2


#------------------------------------------------------------------------------


def cmd_services(opts, args, overDict):
    if len(args) < 2 or args[1] in ['list', 'ls']:

        def _services_update(result):
            for i in range(len(result['result'])):
                r = result['result'][i]
                r['enabled_label'] = 'ENABLED' if r['enabled'] else 'DISABLED'
                r['num_depends'] = len(r['depends'])
                result['result'][i] = r
            return result

        tpl = jsontemplate.Template(templ.TPL_SERVICES)
        return call_websocket_method_transform_template_and_stop('services_list', tpl, _services_update)

    if len(args) == 2 and args[1] in ['tree', 'tr']:

        def _services_update(result):
            for i in range(len(result['result'])):
                r = result['result'][i]
                r['enabled_label'] = 'ENABLED' if r['enabled'] else 'DISABLED'
                r['depends'] = ', '.join(r['depends']) if r['depends'] else 'no dependencies'
                r['offset'] = '| '*r['root_distance']
                result['result'][i] = r
            return result

        tpl = jsontemplate.Template(templ.TPL_SERVICES_TREE)
        return call_websocket_method_transform_template_and_stop('services_list', tpl, _services_update, as_tree=True)

    if len(args) >= 3 and args[1] in ['start', 'enable', 'on']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('service_start', tpl, service_name=args[2])

    if len(args) >= 3 and args[1] in ['stop', 'disable', 'off']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('service_stop', tpl, service_name=args[2])

    if len(args) >= 2:
        tpl = jsontemplate.Template(templ.TPL_SERVICE_INFO)
        return call_websocket_method_template_and_stop('service_info', tpl, service_name=args[1])

    return 2


#------------------------------------------------------------------------------


def cmd_message(opts, args, overDict):
    from twisted.internet import reactor  # @UnresolvedImport
    #     if len(args) < 2 or args[1] == 'list':
    #         tpl = jsontemplate.Template(templ.TPL_RAW)
    #         return call_websocket_method_template_and_stop('list_messages', tpl)
    if len(args) >= 4 and args[1] in ['send', 'to']:
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('message_send', tpl, recipient_id=args[2], data=args[3])

    if len(args) < 2 or args[1] in ['listen', 'read']:
        from bitdust.chat import terminal_chat

        def _send_message(to, msg):
            return call_websocket_method('message_send', recipient_id=to, data=msg)

        def _search_user(inp):
            return call_websocket_method('user_search', nickname=inp)

        terminal_chat.init(
            do_send_message_func=_send_message,
            do_search_user_func=_search_user,
        )
        errors = []

        def _stop(x=None):
            reactor.callInThread(terminal_chat.stop)  # @UndefinedVariable
            reactor.stop()  # @UndefinedVariable
            return True

        def _error(x):
            if str(x).count('ResponseNeverReceived'):
                return x
            errors.append(str(x))
            _stop()
            return x

        def _consume(x=None):
            if x:
                if x['status'] != 'OK':
                    if 'errors' in x:
                        errors.extend(x['errors'])
                    _stop()
                    return x
                for msg in x['result']:
                    terminal_chat.on_incoming_message(msg)

            d = call_websocket_method('message_receive', consumer_callback_id='terminal_chat')
            d.addCallback(_consume)
            d.addErrback(_error)
            return x

        _consume()
        reactor.callInThread(terminal_chat.run)  # @UndefinedVariable
        reactor.run()  # @UndefinedVariable
        terminal_chat.shutdown()
        if len(errors):
            print('\n'.join(errors))
        return 0

    return 2


#------------------------------------------------------------------------------


def cmd_friend(opts, args, overDict):
    tpl_lookup = jsontemplate.Template(templ.TPL_FRIEND_LOOKUP)
    tpl_add = jsontemplate.Template(templ.TPL_RAW)
    if len(args) < 2:
        tpl = jsontemplate.Template(templ.TPL_FRIEND_LOOKUP_REPEATED_SECTION)
        return call_websocket_method_template_and_stop('list_correspondents', tpl)
    elif len(args) > 2 and args[1] in ['check', 'nick', 'nickname', 'test']:
        return call_websocket_method_template_and_stop('user_search', tpl_lookup, nickname=strng.text_type(args[2]))
    elif len(args) > 2 and args[1] in ['add', 'append']:
        inp = strng.text_type(args[2])
        if inp.startswith('http://'):
            return call_websocket_method_template_and_stop('friend_add', tpl_add, trusted_user_id=inp)

        def _lookup(result):
            try:
                if result['result'] == 'exist':
                    print_template(result, tpl_lookup)
                    d = call_websocket_method('friend_add', trusted_user_id=result['idurl'])
                    d.addCallback(print_template_and_stop, tpl_add)
                    d.addErrback(fail_and_stop)
                    return 0
                else:
                    return print_template_and_stop(result, tpl_lookup)
            except:
                from bitdust.logs import lg
                lg.exc()
                return 0

        d = call_websocket_method('user_search', nickname=inp)
        d.addCallback(_lookup)
        d.addErrback(fail_and_stop)
        reactor.run()  # @UndefinedVariable
        return 0
    return 2


#------------------------------------------------------------------------------


def cmd_dhtseed(opts, args, overDict):
    from bitdust.lib import misc
    from bitdust.system import bpio
    from bitdust.main import initializer
    from bitdust.main import shutdowner
    initializer.init_settings()

    if len(args) > 1 and args[1] in ['daemon', 'background', 'detach', 'spawn']:
        appList = bpio.find_main_process()
        if len(appList) > 0:
            print_text('main BitDust process already started: %s' % str(appList))
            shutdowner.shutdown_settings()
            return 0
        print_text('starting Distributed Hash Table seed node and detach main BitDust process')
        result = misc.DoRestart(
            param='dhtseed',
            detach=True,
        )
        try:
            result = result.pid
        except:
            result = str(result)
        shutdowner.shutdown_settings()
        return 0

    from bitdust.dht import dht_service
    from bitdust.logs import lg
    initializer.init_settings()
    dht_service.main(args=args[1:])
    shutdowner.shutdown_settings()
    return 0


#------------------------------------------------------------------------------


def run(opts, args, pars=None, overDict=None, executablePath=None):
    cmd = ''
    if len(args) > 0:
        cmd = args[0].lower()

    from bitdust.system import bpio
    bpio.init()

    #---install---
    if cmd in ['deploy', 'install', 'venv', 'virtualenv']:
        return cmd_deploy(opts, args, overDict)

    #---start---
    if not cmd or cmd == 'start' or cmd == 'go' or cmd == 'run':
        appList = bpio.find_main_process()
        if len(appList) > 0:
            print_text('BitDust already started, found another process: %s' % str(appList))
            return 0
        return run_now(opts, args)

    #---detach---
    elif cmd == 'detach':
        appList = bpio.find_main_process()
        if len(appList) > 0:
            print_text('main BitDust process already started: %s' % str(appList))
            return 0
        from bitdust.lib import misc
        print_text('run and detach main BitDust process')
        result = misc.DoRestart(detach=True, )
        try:
            result = result.pid
        except:
            result = str(result)
        print_text(result)
        return 0

    #---restart---
    elif cmd == 'restart' or cmd == 'reboot':
        appList = bpio.find_main_process()
        if len(appList) == 0:
            return run_now(opts, args)
        ui = False
        if cmd == 'restart':
            ui = True
        print_text('found main BitDust process: %s, executing "api.process_stop()" via WebSocket ... ' % str(appList))

        def done(x):
            print_text('DONE\n', '')
            from twisted.internet import reactor  # @UnresolvedImport
            if reactor.running and not reactor._stopped:  # @UndefinedVariable
                reactor.stop()  # @UndefinedVariable

        def failed(x):
            print_text('soft restart failed, trying to kill running process and force the restart')
            try:
                kill()
            except:
                print_exception()
            from twisted.internet import reactor  # @UnresolvedImport
            reactor.addSystemEventTrigger(  # @UndefinedVariable
                'after',
                'shutdown',
                misc.DoRestart,
                param='show' if ui else '',
                detach=True,
            )
            reactor.stop()  # @UndefinedVariable

        try:
            from twisted.internet import reactor  # @UnresolvedImport
            call_websocket_method('process_restart', websocket_timeout=5).addCallbacks(done, failed)
            reactor.run()  # @UndefinedVariable
        except:
            print_exception()
            return 1

        return 0

    #---show---
    elif cmd == 'show' or cmd == 'open':
        appList_bpgui = bpio.find_process(['bpgui.exe', 'bpgui.py'])
        appList = bpio.find_main_process()
        if len(appList_bpgui) > 0:
            if len(appList) == 0:
                for pid in appList_bpgui:
                    bpio.kill_process(pid)
            else:
                print_text('BitDust GUI already opened, found another process: %s' % str(appList))
                return 0
        if len(appList) == 0:
            print_text('run and detach main BitDust process')
            result = misc.DoRestart('show', detach=True)
            try:
                result = result.pid
            except:
                pass
            print_text(result)
            return 0
        # print_text('found main BitDust process: %s, sending command "show" to start the GUI\n' % str(appList))
        # call_websocket_method('show')
        return 0

    #---stop---
    elif cmd == 'stop' or cmd == 'kill' or cmd == 'shutdown':
        appList = bpio.find_main_process()
        if appList:
            print_text('found main BitDust process: %r ... ' % appList, '')
            try:
                from twisted.internet import reactor  # @UnresolvedImport
                call_websocket_method('process_stop', websocket_timeout=5).addBoth(wait_then_kill)
                reactor.run()  # @UndefinedVariable
                return 0
            except:
                print_exception()
                ret = kill()
                return ret
        else:
            appListAllChilds = bpio.find_main_process(
                check_processid_file=False,
                extra_lookups=[],
            )
            if appListAllChilds:
                print_text('found child BitDust processes: %s, perform "kill process" action ... ' % str(appList), '')
                ret = kill()
                return ret

            print_text('BitDust is not running at the moment')
            return 0

    #---help---
    elif cmd in ['help', 'h', 'hlp', '?']:
        from bitdust.main import help
        if len(args) >= 2 and args[1].lower() == 'schedule':
            print_text(help.schedule_format())
        elif len(args) >= 2 and args[1].lower() == 'settings':
            from bitdust.main import config
            for k in config.conf().listAllEntries():
                print(k, config.conf().getData(k))
        else:
            print_text(help.help_text())
            print_text(pars.format_option_help())
        return 0

    appList = bpio.find_main_process()
    running = (len(appList) > 0)

    overDict = override_options(opts, args)

    #---identity---
    if cmd in ['identity', 'id', 'idurl', 'globalid', 'globid', 'glid', 'gid']:
        return cmd_identity(opts, args, overDict, running, executablePath)

    #---key---
    elif cmd in ['key', 'keys', 'pk', 'private_key', 'priv']:
        return cmd_key(opts, args, overDict, running, executablePath)

    #---ping---
    if cmd == 'ping' or cmd == 'call' or cmd == 'sendid':
        if len(args) < 1:
            return 2
        tpl = jsontemplate.Template(templ.TPL_RAW)
        return call_websocket_method_template_and_stop('user_ping', tpl, user_id=args[1])

    #---config---
    elif cmd in ['set', 'get', 'conf', 'config', 'option', 'setting']:
        if len(args) == 1 or args[1].lower() in ['help', '?']:
            from bitdust.main import help
            print_text(help.settings_help())
            return 0
        if not running:
            return cmd_set(opts, args, overDict)
        return cmd_set_request(opts, args, overDict)

    #---reconnect---
    if cmd in ['reconnect', 'rejoin', 'connect']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_reconnect(opts, args, overDict)

    #---api---
    elif cmd in ['api', 'call']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_api(opts, args, overDict, executablePath)

    #---messages---
    elif cmd in ['msg', 'message', 'messages', 'chat', 'talk']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_message(opts, args, overDict)

    #---suppliers---
    elif cmd in ['suppliers', 'supplier', 'sup', 'supp', 'sp']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_suppliers(opts, args, overDict)

    #---customers---
    elif cmd in ['customers', 'customer', 'cus', 'cust', 'cu']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_customers(opts, args, overDict)

    #---storage---
    elif cmd in ['storage', 'space']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_storage(opts, args, overDict)

    #---storage---
    elif cmd in ['network', 'net', 'nw']:
        return cmd_network(opts, args, overDict, running)

    #---automats---
    elif cmd in ['st', 'state', 'automats', 'aut', 'states', 'machines']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_automats(opts, args, overDict)

    #---services---
    elif cmd in ['services', 'service', 'svc', 'serv', 'srv']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_services(opts, args, overDict)

    #---friends---
    elif cmd == 'friend' or cmd == 'friends' or cmd == 'buddy' or cmd == 'correspondent' or cmd == 'contact' or cmd == 'peer':
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_friend(opts, args, overDict)

    #---file---
    elif cmd in ['file', 'files', 'fi', 'fs', 'f', 'folder', 'dir']:
        if not running:
            print_text('BitDust is not running at the moment\n')
            return 0
        return cmd_file(opts, args, overDict, executablePath)

    #---dhtseed---
    elif cmd == 'dhtseed':
        appList = bpio.find_main_process(
            check_processid_file=False,
            extra_lookups=[],
        )
        running = (len(appList) > 0)
        if running:
            print_text('BitDust is running at the moment, need to stop the software first\n')
            return 0
        return cmd_dhtseed(opts, args, overDict)

    #---version---
    elif cmd in ['version', 'v', 'ver']:
        from bitdust.main import settings
        from bitdust.lib import misc
        ver = bpio.ReadTextFile(settings.VersionNumberFile()).strip()
        chksum = bpio.ReadTextFile(settings.CheckSumFile()).strip()
        # repo, location = misc.ReadRepoLocation()
        print_text('checksum:   %s' % chksum)
        print_text('version:    %s' % ver)
        # print_text('repository: %s' % repo)
        # print_text('location:   %s' % location)
        return 0

    return 2


#------------------------------------------------------------------------------


def main():
    try:
        from bitdust.system import bpio
    except:
        dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
        sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
        from distutils.sysconfig import get_python_lib
        sys.path.append(os.path.join(get_python_lib(), 'bitdust'))
        try:
            from bitdust.system import bpio
        except:
            print_text('ERROR! can not import working code.  Python Path:\n')
            print_text('\n'.join(sys.path))
            return 1

    pars = parser()
    (opts, args) = pars.parse_args()

    if opts.verbose:
        print_copyright()

    return run(opts, args)


#------------------------------------------------------------------------------



#!/usr/bin/python
#geth_service.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (geth_service.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: geth_service

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import os
import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in geth_service.py')

from twisted.internet import protocol

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import settings

from bitdust.updates import git_proc

#------------------------------------------------------------------------------


def init():
    lg.out(4, 'geth_service.init')
    try:
        verify_local_install()
    except ValueError:
        deploy(callback=run_geth_node)
    else:
        run_geth_node()


def shutdown():
    lg.out(4, 'geth_service.shutdown')


#------------------------------------------------------------------------------


def verify_local_install():
    ethereum_location = os.path.join(settings.AppDataDir(), 'ethereum')
    if not os.path.isdir(ethereum_location):
        raise ValueError('Ethereum root location not found: {}'.format(ethereum_location))
    geth_location = os.path.join(ethereum_location, 'go-ethereum')
    if not os.path.isdir(ethereum_location):
        raise ValueError('Ethereum geth process folder not found: {}'.format(ethereum_location))
    geth_bin_path = os.path.join(geth_location, 'build', 'bin', 'geth')
    if not os.path.isfile(geth_bin_path):
        raise ValueError('Ethereum geth process executable not found: {}'.format(geth_bin_path))
    return True


def verify_global_install():
    # TODO: try to run "geth" in shell
    return True


#------------------------------------------------------------------------------


def clone(callback=None):
    ethereum_location = os.path.join(settings.AppDataDir(), 'ethereum')
    if not os.path.isdir(ethereum_location):
        os.makedirs(ethereum_location)
    geth_location = os.path.join(ethereum_location, 'go-ethereum')
    if os.path.exists(geth_location):
        bpio.rmdir_recursive(geth_location)
    git_proc.run(
        [
            'clone',
            '--verbose',
            '--depth',
            '1',
            'https://github.com/ethereum/go-ethereum',
            geth_location,
        ],
        base_dir=ethereum_location,
        env=os.environ,
        callback_func=callback,
    )


def make(callback=None):
    geth_location = os.path.join(settings.AppDataDir(), 'ethereum', 'go-ethereum')
    execute(['make', 'geth'], base_dir=geth_location, env=os.environ, callback=callback)


def deploy(callback=None):
    def _clone(out, retcode):
        make(callback=callback)

    clone(callback=_clone)


#------------------------------------------------------------------------------


def run(cmdargs, callback=None):
    geth_location = os.path.join(
        settings.AppDataDir(),
        'ethereum',
        'go-ethereum',
    )
    cmdargs = [
        'build/bin/geth',
    ] + cmdargs
    execute(cmdargs, base_dir=geth_location, env=os.environ, callback=callback)


def run_geth_node():
    geth_datadir = os.path.join(settings.AppDataDir(), 'ethereum', 'datadir')
    if not os.path.isdir(geth_datadir):
        os.makedirs(geth_datadir)
    run(['--datadir="{}"'.format(geth_datadir), '--verbosity', '4', '--ipcdisable', '--port', '30300', '--rpcport', '8100', '--networkid', '1'])


#------------------------------------------------------------------------------


def execute(cmdargs, base_dir=None, process_protocol=None, env=None, callback=None):
    global _CurrentProcess
    if _Debug:
        lg.out(_DebugLevel, 'geth_service.execute: "%s" in %s' % (' '.join(cmdargs), base_dir))
    executable = cmdargs[0]
    if bpio.Windows():
        from twisted.internet import _dumbwin32proc
        real_CreateProcess = _dumbwin32proc.win32process.CreateProcess

        def fake_createprocess(_appName, _commandLine, _processAttributes, _threadAttributes, _bInheritHandles, creationFlags, _newEnvironment, _currentDirectory, startupinfo):
            import win32con
            import subprocess
            flags = win32con.CREATE_NO_WINDOW
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            startupinfo.wShowWindow = subprocess.SW_HIDE
            return real_CreateProcess(_appName, _commandLine, _processAttributes, _threadAttributes, _bInheritHandles, flags, _newEnvironment, _currentDirectory, startupinfo)

        setattr(_dumbwin32proc.win32process, 'CreateProcess', fake_createprocess)

    if process_protocol is None:
        process_protocol = GethProcessProtocol(callback)
    try:
        _CurrentProcess = reactor.spawnProcess(process_protocol, executable, cmdargs, path=base_dir, env=env)
    except:
        lg.exc()
        return None
    if bpio.Windows():
        setattr(_dumbwin32proc.win32process, 'CreateProcess', real_CreateProcess)
    return _CurrentProcess


#------------------------------------------------------------------------------


class GethProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, callback):
        """
        """
        self.callback = callback
        self.out = ''
        self.err = ''

    def errReceived(self, inp):
        """
        """
        self.err += inp
        for line in inp.splitlines():
            if _Debug:
                lg.out(_DebugLevel, '%s' % line)

    def outReceived(self, inp):
        """
        """
        self.out += inp
        for line in inp.splitlines():
            if _Debug:
                lg.out(_DebugLevel, '%s' % line)

    def processEnded(self, reason):
        """
        """
        if _Debug:
            lg.out(_DebugLevel, 'geth process FINISHED : %s' % reason.value.exitCode)
        if self.callback:
            self.callback(self.out, reason.value.exitCode)


#------------------------------------------------------------------------------

if __name__ == '__main__':
    bpio.init()
    lg.set_debug_level(18)
    reactor.callWhenRunning(run_geth_node)
    reactor.run()



#!/usr/bin/env python
# blockchain_registrator.py
#
"""
.. module:: blockchain_registrator
.. role:: red

BitDust blockchain_registrator() Automat

EVENTS:
    * :red:`dht-read-failed`
    * :red:`dht-read-success`
    * :red:`dht-write-failed`
    * :red:`dht-write-success`
    * :red:`record-already-exist`
    * :red:`shutdown`
    * :red:`start`
    * :red:`timer-5min`
    * :red:`tx-not-found`
    * :red:`valid-tx-found`
"""

#------------------------------------------------------------------------------

import random

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import strng

from bitdust.blockchain import bismuth_wallet

from bitdust.dht import dht_records

from bitdust.userid import my_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

_BlockchainRegistrator = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with `blockchain_registrator()` machine.
    """
    global _BlockchainRegistrator
    if event is None:
        return _BlockchainRegistrator
    if _BlockchainRegistrator is None:
        if event == 'shutdown':
            return _BlockchainRegistrator
        # TODO: set automat name and starting state here
        _BlockchainRegistrator = BlockchainRegistrator(
            name='blockchain_registrator',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
            publish_events=False,
        )
    if event is not None:
        _BlockchainRegistrator.automat(event, *args, **kwargs)
    return _BlockchainRegistrator


class BlockchainRegistrator(automat.Automat):

    """
    This class implements all the functionality of ``blockchain_registrator()`` state machine.
    """

    timers = {
        'timer-5min': (300, ['BLOCKCHAIN_READ']),
    }

    def __init__(self, debug_level=0, log_events=False, log_transitions=False, publish_events=False, **kwargs):
        """
        Builds `blockchain_registrator()` state machine.
        """
        self.dht_position = 1
        super(BlockchainRegistrator, self).__init__(
            debug_level=debug_level,
            log_events=log_events,
            log_transitions=log_transitions,
            publish_events=publish_events,
            **kwargs,
        )

    def __repr__(self):
        return '%s[%d](%s)' % (self.id, self.dht_position, self.state)

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `blockchain_registrator()` machine.
        """

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `blockchain_registrator()` state were changed.
        """

    def state_not_changed(self, curstate, event, *args, **kwargs):
        """
        This method intended to catch the moment when some event was fired in the `blockchain_registrator()`
        but automat state was not changed.
        """

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <https://github.com/vesellov/visio2python>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'start':
                self.state = 'BLOCKCHAIN_READ'
                self.doInit(*args, **kwargs)
                self.doSearchTransaction(*args, **kwargs)
        #---DHT_READ---
        elif self.state == 'DHT_READ':
            if event == 'dht-read-failed':
                self.state = 'DHT_WRITE'
                self.doDHTWrite(*args, **kwargs)
            elif event == 'dht-read-success':
                self.state = 'BLOCKCHAIN_READ'
            elif event == 'record-already-exist':
                self.doNextPosition(*args, **kwargs)
                self.doDHTRead(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---DHT_WRITE---
        elif self.state == 'DHT_WRITE':
            if event == 'dht-write-success' or (event == 'dht-write-failed' and self.Attempts > 5):
                self.state = 'BLOCKCHAIN_READ'
                self.Attempts = 0
            elif event == 'dht-write-failed' and self.Attempts <= 5:
                self.state = 'DHT_READ'
                self.Attempts += 1
                self.doNextPosition(*args, **kwargs)
                self.doDHTRead(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---BLOCKCHAIN_READ---
        elif self.state == 'BLOCKCHAIN_READ':
            if event == 'tx-not-found':
                self.state = 'DHT_READ'
                self.Attempts = 0
                self.doDHTRead(*args, **kwargs)
            elif event == 'valid-tx-found':
                self.state = 'READY'
                self.doDHTErase(*args, **kwargs)
                self.doReportReady(*args, **kwargs)
            elif event == 'timer-5min':
                self.doSearchTransaction(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---READY---
        elif self.state == 'READY':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.result_defer = kwargs.get('result_defer')
        self.dht_position = 1
        self.dht_read_defer = None

    def doNextPosition(self, *args, **kwargs):
        """
        Action method.
        """
        current_position = self.dht_position
        max_position = self.dht_position*2 + 1
        self.dht_position += random.randint(current_position + 1, max_position)
        if self.dht_position > 500:
            self.dht_position = 1

    def doSearchTransaction(self, *args, **kwargs):
        """
        Action method.
        """
        clean_public_key = strng.to_text(my_id.getLocalIdentity().getPublicKey())[:10000].replace('ssh-rsa ', '')
        results = bismuth_wallet.find_transaction(
            recipient=bismuth_wallet.my_wallet_address(),
            operation='identity',
            openfield='{}:{}'.format(my_id.getIDName(), clean_public_key),
        )
        if results:
            self.automat('valid-tx-found')
        else:
            self.automat('tx-not-found')

    def doDHTRead(self, *args, **kwargs):
        """
        Action method.
        """
        if self.dht_read_defer is not None:
            self.dht_read_defer.pause()
            self.dht_read_defer.cancel()
            self.dht_read_defer = None
        self.dht_read_defer = dht_records.get_bismuth_identity_request(position=self.dht_position, use_cache=False)
        self.dht_read_defer.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='blockchain_registrator.doDHTReadKey')
        self.dht_read_defer.addCallback(self._dht_read_result, self.dht_position)
        self.dht_read_defer.addErrback(self._dht_read_failed)

    def doDHTWrite(self, *args, **kwargs):
        """
        Action method.
        """
        clean_public_key = strng.to_text(my_id.getLocalIdentity().getPublicKey())[:10000].replace('ssh-rsa ', '')
        d = dht_records.set_bismuth_identity_request(
            position=self.dht_position,
            idurl=my_id.getIDURL(),
            public_key=clean_public_key,
            wallet_address=bismuth_wallet.my_wallet_address(),
        )
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='blockchain_registrator.doDHTWriteKey')
        d.addCallback(self._dht_write_result)
        d.addErrback(lambda err: self.automat('dht-write-failed', err))

    def doDHTErase(self, *args, **kwargs):
        """
        Action method.
        """
        dht_records.erase_bismuth_identity_request(position=self.dht_position)

    def doReportReady(self, *args, **kwargs):
        """
        Action method.
        """
        if self.result_defer:
            self.result_defer.callback(True)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()
        global _BlockchainRegistrator
        del _BlockchainRegistrator
        _BlockchainRegistrator = None

    def _dht_read_result(self, value, position):
        if _Debug:
            lg.args(_DebugLevel, position=position, value=value)
        self.dht_read_defer = None
        if not value:
            self.automat('dht-read-failed')
            return
        try:
            idurl_at_position = id_url.field(value['idurl'])
        except:
            lg.exc()
            self.automat('dht-read-failed')
            return
        if id_url.is_the_same(idurl_at_position, my_id.getIDURL()):
            if value['wallet_address'] != bismuth_wallet.my_wallet_address():
                self.automat('dht-read-failed')
            else:
                self.automat('dht-read-success', value)
        else:
            self.automat('record-already-exist', value)

    def _dht_read_failed(self, result):
        if _Debug:
            lg.args(_DebugLevel, result=result)
        self.dht_read_defer = None
        self.automat('dht-read-failed', result)

    def _dht_write_result(self, nodes):
        if _Debug:
            lg.args(_DebugLevel, nodes=nodes)
        if len(nodes) > 0:
            self.automat('dht-write-success')
        else:
            self.automat('dht-write-failed', nodes)



#!/usr/bin/env python
# bptester.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (bptester.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: bptester.

This is a BitDust child process, do monitoring of customer's files.

In case some of customers do not play fair - need to stop this.:

    * check if he use more space than we gave him, remove too old files
    * test/remove files after list of customers was changed
    * check all packets to be valid
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

import os
import sys
import time
from io import open

#------------------------------------------------------------------------------

AppData = ''
CurrentNetwork = ''

#------------------------------------------------------------------------------


def sharedPath(filename, subdir='logs'):
    global AppData
    global CurrentNetwork
    if not AppData:
        curdir = os.getcwd()  # os.path.dirname(os.path.abspath(sys.executable))
        if os.path.isfile(os.path.join(curdir, 'appdata')):
            try:
                appdata = os.path.abspath(open(os.path.join(curdir, 'appdata'), 'r').read().strip())
            except:
                appdata = os.path.join(os.path.expanduser('~'), '.bitdust')
            if not os.path.isdir(appdata):
                appdata = os.path.join(os.path.expanduser('~'), '.bitdust')
        else:
            if sys.executable == 'android_python' or ('ANDROID_ARGUMENT' in os.environ or 'ANDROID_ROOT' in os.environ):
                from android.storage import app_storage_path  # @UnresolvedImport
                appdata = os.path.join(app_storage_path(), '.bitdust')
            else:
                appdata = os.path.join(os.path.expanduser('~'), '.bitdust')
        AppData = appdata
    if not CurrentNetwork:
        try:
            cur_network = open(os.path.join(AppData, 'current_network'), 'r').read().strip()
        except:
            cur_network = 'default'
        if not os.path.isdir(os.path.join(AppData, cur_network)):
            cur_network = 'default'
        CurrentNetwork = cur_network
    return os.path.join(AppData, CurrentNetwork, subdir, filename)


def logfilepath():
    """
    A file path to the file where ``bptester`` will write logs.

    Need to make sure the ``bptester`` log is in a directory the user
    has permissions for, Such as the customer data directory.  Possibly
    move to temp directory?
    """
    return sharedPath('bptester.log')


def printlog(txt):
    """
    Write a line to the log file.
    """
    try:
        if sys.version_info[0] == 3:
            if not isinstance(txt, str):
                txt = txt.decode()
        else:
            if not isinstance(txt, unicode):  # @UndefinedVariable
                txt = txt.decode()
        lf = open(logfilepath(), 'a')
        lf.write(txt + '\n')
        lf.close()
    except:
        pass


#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.lib import misc

from bitdust.crypt import signed

from bitdust.storage import accounting

from bitdust.main import settings

from bitdust.userid import global_id
from bitdust.userid import id_url

#-------------------------------------------------------------------------------


def SpaceTime():
    """
    Test all packets for each customer.

    Check if he use more space than we gave him and if packets is too
    old.
    """
    if _Debug:
        printlog('SpaceTime %r' % time.strftime('%a, %d %b %Y %H:%M:%S +0000'))
    space, _ = accounting.read_customers_quotas()
    if space is None:
        if _Debug:
            printlog('SpaceTime ERROR customers quotas file can not be read or it is empty, skip')
        return False
    customers_dir = settings.getCustomersFilesDir()
    if not os.path.exists(customers_dir):
        if _Debug:
            printlog('SpaceTime ERROR customers folder not exist: %r' % customers_dir)
        return False
    remove_list = {}
    used_space = accounting.read_customers_usage()
    for customer_filename in os.listdir(customers_dir):
        onecustdir = os.path.join(customers_dir, customer_filename)
        if not os.path.isdir(onecustdir):
            remove_list[onecustdir] = 'is not a folder'
            continue
        # idurl = nameurl.FilenameUrl(customer_filename)
        idurl = global_id.GlobalUserToIDURL(customer_filename)
        if idurl is None:
            remove_list[onecustdir] = 'wrong folder name'
            continue
        curspace = space.get(idurl.to_bin(), None)
        if curspace is None:
            remove_list[onecustdir] = 'not found in space file'
            continue
        try:
            maxspaceV = int(curspace)
        except:
            remove_list[onecustdir] = 'wrong space value'
            continue
        timedict = {}
        sizedict = {}

        def cb(path, subpath, name):
            if not os.path.isfile(path):
                return True
            stats = os.stat(path)
            timedict[path] = stats.st_ctime
            sizedict[path] = stats.st_size
            return False

        for key_alias in os.listdir(onecustdir):
            if not misc.ValidKeyAlias(key_alias):
                remove_list[onecustdir] = 'invalid key alias'
                continue
            okekeydir = os.path.join(onecustdir, key_alias)
            bpio.traverse_dir_recursive(cb, okekeydir)
            currentV = 0
            for path in sorted(list(timedict.keys()), key=lambda x: timedict[x], reverse=True):
                filesize = sizedict.get(path, 0)
                currentV += filesize
                if currentV < maxspaceV:
                    continue
                try:
                    os.remove(path)
                    if _Debug:
                        printlog('SpaceTime %r file removed (cur:%s, max: %s)' % (path, str(currentV), str(maxspaceV)))
                except:
                    if _Debug:
                        printlog('SpaceTime ERROR removing %r' % path)
                # time.sleep(0.01)

        used_space[idurl.to_bin()] = str(currentV)
        timedict.clear()
        sizedict.clear()

    for customer_idurl_bin in list(used_space.keys()):
        if not id_url.field(customer_idurl_bin).is_latest():
            latest_customer_idurl_bin = id_url.field(customer_idurl_bin).to_bin()
            if latest_customer_idurl_bin != customer_idurl_bin:
                used_space[latest_customer_idurl_bin] = used_space.pop(customer_idurl_bin)
                if _Debug:
                    printlog('found customer idurl rotated in customer usage dictionary : %r -> %r' % (
                        latest_customer_idurl_bin,
                        customer_idurl_bin,
                    ))

    for path in remove_list.keys():
        if not os.path.exists(path):
            continue
        if os.path.isdir(path):
            try:
                bpio._dir_remove(path)
                if _Debug:
                    printlog('SpaceTime %r dir removed (%s)' % (path, remove_list[path]))
            except:
                if _Debug:
                    printlog('SpaceTime ERROR removing %r' % path)
            continue
        try:
            if not os.access(path, os.W_OK):
                os.chmod(path, 0o600)
        except:
            pass
        try:
            os.remove(path)
            if _Debug:
                printlog('SpaceTime %r file removed (%s)' % (path, remove_list[path]))
        except:
            if _Debug:
                printlog('SpaceTime ERROR removing %r' % path)
    del remove_list

    accounting.update_customers_usage(used_space)

    return True


#------------------------------------------------------------------------------


def UpdateCustomers():
    """
    Test packets after list of customers was changed.
    """
    space, _ = accounting.read_customers_quotas()
    if space is None:
        if _Debug:
            printlog('UpdateCustomers ERROR space file can not be read')
        return False
    customers_dir = settings.getCustomersFilesDir()
    if not os.path.exists(customers_dir):
        if _Debug:
            printlog('UpdateCustomers ERROR customers folder not exist')
        return False

    remove_list = {}
    for customer_filename in os.listdir(customers_dir):
        onecustdir = os.path.join(customers_dir, customer_filename)
        if not os.path.isdir(onecustdir):
            remove_list[onecustdir] = 'is not a folder'
            continue
        # idurl = nameurl.FilenameUrl(customer_filename)
        idurl = global_id.GlobalUserToIDURL(customer_filename)
        if idurl is None:
            remove_list[onecustdir] = 'wrong folder name'
            continue
        curspace = space.get(idurl.to_bin(), None)
        if curspace is None:
            remove_list[onecustdir] = 'is not a customer'
            continue

    for path in remove_list.keys():
        if not os.path.exists(path):
            continue
        if os.path.isdir(path):
            try:
                bpio._dir_remove(path)
                if _Debug:
                    printlog('UpdateCustomers %r folder removed (%s)' % (
                        path,
                        remove_list[path],
                    ))
            except:
                if _Debug:
                    printlog('UpdateCustomers ERROR removing %r' % path)
            continue
        try:
            if not os.access(path, os.W_OK):
                os.chmod(path, 0o600)
        except:
            pass
        try:
            os.remove(path)
            if _Debug:
                printlog('UpdateCustomers %r file removed (%s)' % (
                    path,
                    remove_list[path],
                ))
        except:
            if _Debug:
                printlog('UpdateCustomers ERROR removing %r' % path)
    if _Debug:
        printlog('UpdateCustomers %r' % time.strftime('%a, %d %b %Y %H:%M:%S +0000'))
    return True


#------------------------------------------------------------------------------


def Validate():
    """
    Check all packets to be valid.
    """
    if _Debug:
        printlog('Validate %r' % time.strftime('%a, %d %b %Y %H:%M:%S +0000'))
    customers_dir = settings.getCustomersFilesDir()
    if not os.path.exists(customers_dir):
        return False

    for customer_filename in os.listdir(customers_dir):
        onecustdir = os.path.join(customers_dir, customer_filename)
        if not os.path.isdir(onecustdir):
            continue
        for key_alias_filename in os.listdir(onecustdir):
            onekeydir = os.path.join(onecustdir, key_alias_filename)
            if not os.path.isdir(onekeydir):
                continue

            def cb(path, subpath, name):
                if not os.path.isfile(path):
                    return True
                packetsrc = bpio.ReadBinaryFile(path)
                if not packetsrc:
                    try:
                        os.remove(path)  # if is is no good it is of no use to anyone
                        if _Debug:
                            printlog('Validate %r removed (empty file)' % path)
                    except:
                        if _Debug:
                            printlog('Validate ERROR removing %r' % path)
                        return False
                p = signed.Unserialize(packetsrc)
                if p is None:
                    try:
                        os.remove(path)  # if is is no good it is of no use to anyone
                        if _Debug:
                            printlog('Validate %r removed (unserialize error)' % path)
                    except:
                        if _Debug:
                            printlog('Validate ERROR removing %r')
                        return False
                result = p.Valid()
                packetsrc = ''
                del p
                if not result:
                    try:
                        os.remove(path)  # if is is no good it is of no use to anyone
                        if _Debug:
                            printlog('Validate %r removed (invalid packet)' % path)
                    except:
                        if _Debug:
                            printlog('Validate ERROR removing %r' % path)
                        return False
                time.sleep(0.1)
                return False

            bpio.traverse_dir_recursive(cb, onekeydir)

    return True


#------------------------------------------------------------------------------


def main():
    """
    Entry point.
    """
    if len(sys.argv) < 2:
        return
    bpio.init()
    lg.disable_logs()
    lg.disable_output()
    settings.init()
    lg.set_debug_level(0)
    id_url.init()
    commands = {
        'update_customers': UpdateCustomers,
        'validate': Validate,
        'space_time': SpaceTime,
    }
    cmd = commands.get(sys.argv[1], None)
    if not cmd:
        if _Debug:
            printlog('ERROR wrong command: %r' % sys.argv)
        return
    cmd()
    settings.shutdown()
    id_url.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/python
# backup.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: backup.

.. raw:: html

    <a href="https://bitdust.io/automats/backup/backup.png" target="_blank">
    <img src="https://bitdust.io/automats/backup/backup.png" style="max-width:100%;">
    </a>

The core module.
The ``backup()`` state machine is doing a bunch of things to create a single backup.

Here is an interfaces between a pipe from something like tar
and the twisted code for rest of BitDust.

Main idea:
   1) when a backup is started a backup object is created
   2) get a file descriptor for the process creating the tar archive
   3) always use select/poll before reading, so never block the main process
   4) also poll to see if more data is needed to create a block
   5) number/name blocks so can be sure what is what when we read back later
   6) encrypt the block data into ``encrypted_blocks``
   7) call ``p2p.raidmake`` to split block and make "Parity" packets (pieces of block)
   8) notify the top level code about new pieces of data to send on suppliers

This state machine controls the data read from the folder,
partition the data into blocks,
block encryption using the private key and the transfer of units to the suppliers.

Reading is performed from the opened ".tar" pipe and must be finished
as soon as empty chunk of data were read from the pipe.
The encrypted data blocks are stored in a temporary folder on the HDD
and deleted (user configurable) as soon as the suppliers have them.

For each block must be received delivery report: positive or negative.

Process will be completed as soon as all data will be read from the folder
and all blocks will receive a delivery report.

EVENTS:
    * :red:`block-encrypted`
    * :red:`block-raid-done`
    * :red:`block-raid-started`
    * :red:`fail`
    * :red:`read-success`
    * :red:`start`
    * :red:`timer-001sec`
    * :red:`timer-01sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import BytesIO

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import sys
import time
import gc

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in backup.py')

from twisted.internet.defer import maybeDeferred, Deferred, succeed

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.append(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..'))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import packetid
from bitdust.lib import strng

from bitdust.userid import my_id
from bitdust.userid import global_id

from bitdust.system import nonblocking
from bitdust.system import tmpfile

from bitdust.main import settings
from bitdust.main import events

from bitdust.automats import automat

from bitdust.raid import eccmap
from bitdust.raid import raid_worker

from bitdust.crypt import encrypted
from bitdust.crypt import key

#-------------------------------------------------------------------------------


class backup(automat.Automat):

    """
    A class to run the backup process, data is read from pipe.
    """

    timers = {
        'timer-01sec': (0.1, ['RAID']),
        'timer-001sec': (0.01, ['READ']),
    }

    def __init__(
        self,
        backupID,
        pipe,
        finishCallback=None,
        blockResultCallback=None,
        notifyNewDataCallback=None,
        blockSize=None,
        sourcePath=None,
        keyID=None,
        ecc_map=None,
        creatorIDURL=None,
    ):
        self.backupID = backupID
        self.creatorIDURL = creatorIDURL or my_id.getIDURL()
        _parts = packetid.SplitBackupID(self.backupID)
        self.customerGlobalID = _parts[0]
        self.pathID = _parts[1]
        self.version = _parts[2]
        self.customerIDURL = global_id.GlobalUserToIDURL(self.customerGlobalID)
        self.sourcePath = sourcePath
        self.keyID = keyID
        self.eccmap = ecc_map or eccmap.Current()
        self.pipe = pipe
        self.blockSize = blockSize
        if self.blockSize is None:
            self.blockSize = settings.getBackupBlockSize()
        self.ask4abort = False
        self.terminating = False
        self.stateEOF = False
        self.stateReading = False
        self.closed = False
        self.currentBlockData = BytesIO()
        self.currentBlockSize = 0
        self.workBlocks = {}
        self.blockNumber = 0
        self.dataSent = 0
        self.blocksSent = 0
        self.totalSize = -1
        self.resultDefer = Deferred()
        self.finishCallback = finishCallback
        self.blockResultCallback = blockResultCallback
        self.notifyNewDataCallback = notifyNewDataCallback
        automat.Automat.__init__(
            self,
            name='backup_%s' % self.version,
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )

    def init(self):
        """
        """
        self.log_transitions = _Debug

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'start':
                self.state = 'READ'
                self.doInit(*args, **kwargs)
                self.doFirstBlock(*args, **kwargs)
        #---READ---
        elif self.state == 'READ':
            if event == 'read-success' and not self.isReadingNow(*args, **kwargs) and (self.isBlockReady(*args, **kwargs) or self.isEOF(*args, **kwargs)):
                self.state = 'ENCRYPT'
                self.doEncryptBlock(*args, **kwargs)
            elif event == 'fail' or ((event == 'read-success' or event == 'timer-001sec') and self.isAborted(*args, **kwargs)):
                self.state = 'ABORTED'
                self.doClose(*args, **kwargs)
                self.doReport(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'read-success' or event
                  == 'timer-001sec') and not self.isAborted(*args, **kwargs) and self.isPipeReady(*args, **kwargs) and not self.isEOF(*args, **kwargs) and not self.isReadingNow(*args, **kwargs) and not self.isBlockReady(*args, **kwargs):
                self.doRead(*args, **kwargs)
            elif event == 'block-raid-done' and not self.isAborted(*args, **kwargs):
                self.doPopBlock(*args, **kwargs)
                self.doBlockReport(*args, **kwargs)
                self.doNotifyNewData(*args, **kwargs)
        #---RAID---
        elif self.state == 'RAID':
            if event == 'block-raid-done' and not self.isMoreBlocks(*args, **kwargs) and not self.isAborted(*args, **kwargs):
                self.state = 'DONE'
                self.doPopBlock(*args, **kwargs)
                self.doBlockReport(*args, **kwargs)
                self.doNotifyNewData(*args, **kwargs)
                self.doClose(*args, **kwargs)
                self.doReport(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'block-raid-started' and not self.isEOF(*args, **kwargs) and not self.isAborted(*args, **kwargs):
                self.state = 'READ'
                self.doNextBlock(*args, **kwargs)
                self.doRead(*args, **kwargs)
            elif event == 'block-raid-done' and self.isMoreBlocks(*args, **kwargs) and not self.isAborted(*args, **kwargs):
                self.doPopBlock(*args, **kwargs)
                self.doBlockReport(*args, **kwargs)
                self.doNotifyNewData(*args, **kwargs)
            elif event == 'fail' or ((event == 'timer-01sec' or event == 'block-raid-done' or event == 'block-raid-started') and self.isAborted(*args, **kwargs)):
                self.state = 'ABORTED'
                self.doClose(*args, **kwargs)
                self.doReport(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---ENCRYPT---
        elif self.state == 'ENCRYPT':
            if event == 'block-encrypted':
                self.state = 'RAID'
                self.doBlockPushAndRaid(*args, **kwargs)
            elif event == 'fail':
                self.state = 'ABORTED'
                self.doClose(*args, **kwargs)
                self.doReport(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'block-raid-done' and not self.isAborted(*args, **kwargs):
                self.doPopBlock(*args, **kwargs)
                self.doBlockReport(*args, **kwargs)
                self.doNotifyNewData(*args, **kwargs)
        #---DONE---
        elif self.state == 'DONE':
            pass
        #---ABORTED---
        elif self.state == 'ABORTED':
            pass
        return None

    def isAborted(self, *args, **kwargs):
        """
        Return current value of ``ask4abort`` flag.
        """
        # if _Debug:
        #     lg.args(_DebugLevel, ask4abort=self.ask4abort)
        return self.ask4abort

    def isPipeReady(self, *args, **kwargs):
        """
        Return True if ``pipe`` object exist and is ready for reading a the new
        chunk of data.
        """
        if _Debug:
            lg.args(_DebugLevel, pipe=bool(self.pipe), pipe_state=(self.pipe.state() if self.pipe else None))
        if self.pipe is None:
            lg.warn('pipe is None')
            return False
        return self.pipe.state() in [nonblocking.PIPE_CLOSED, nonblocking.PIPE_READY2READ]

    def isBlockReady(self, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, currentBlockSize=self.currentBlockSize, blockSize=self.blockSize)
        return self.currentBlockSize >= self.blockSize

    def isEOF(self, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, stateEOF=self.stateEOF)
        return self.stateEOF

    def isReadingNow(self, *args, **kwargs):
        if _Debug:
            lg.args(_DebugLevel, stateReading=self.stateReading)
        return self.stateReading

    def isMoreBlocks(self, *args, **kwargs):
        """
        Condition method.
        """
        if _Debug:
            lg.args(_DebugLevel, workBlocks=len(self.workBlocks))
        return len(self.workBlocks) > 1

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        events.send('backup-started', data=dict(backup_id=self.backupID))

    def doRead(self, *args, **kwargs):
        """
        Action method.
        """

        def readChunk():
            size = self.blockSize - self.currentBlockSize
            if size < 0:
                if _Debug:
                    lg.args(_DebugLevel, eccmap_nodes=self.eccmap.nodes(), block_size=self.blockSize, current_block_size=self.currentBlockSize)
                raise Exception('size < 0, blockSize=%s, currentBlockSize=%s' % (self.blockSize, self.currentBlockSize))
            elif size == 0:
                return succeed(b'')
            if self.pipe is None:
                raise Exception('backup.pipe is None')
            if self.pipe.state() == 2:
                if _Debug:
                    lg.out(_DebugLevel, 'backup.readChunk the state is PIPE_CLOSED in %r' % self)
                return succeed(b'')
            if self.pipe.state() == 0:
                if _Debug:
                    lg.out(_DebugLevel, 'backup.readChunk the state is PIPE_EMPTY in %r' % self)
                return succeed(b'')
            if _Debug:
                lg.args(_DebugLevel, size=size)
            return self.pipe.read_defer(size)

        def readDone(data):
            try:
                self.stateReading = False
                self.currentBlockData.write(data)
                self.currentBlockSize += len(data)
            except:
                lg.exc()
                self.automat('fail', None)
                return None
            if not data:
                self.stateEOF = True
            if _Debug:
                lg.out(_DebugLevel, 'backup.readDone %d bytes' % len(data))
            reactor.callLater(0, self.automat, 'read-success')  # @UndefinedVariable
            return data

        def readFailed(err):
            lg.err(err)
            self.automat('fail', err)
            return None

        self.stateReading = True
        d = readChunk()
        d.addCallback(readDone)
        d.addErrback(readFailed)

    def doEncryptBlock(self, *args, **kwargs):
        """
        Action method.
        """

        def _doBlock():
            dt = time.time()
            raw_bytes = self.currentBlockData.getvalue()
            block = encrypted.Block(
                CreatorID=self.creatorIDURL,
                BackupID=self.backupID,
                BlockNumber=self.blockNumber,
                SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
                SessionKeyType=key.SessionKeyType(),
                LastBlock=self.stateEOF,
                Data=raw_bytes,
                EncryptKey=self.keyID,
            )
            del raw_bytes
            if _Debug:
                lg.out(_DebugLevel, 'backup.doEncryptBlock blockNumber=%d size=%d atEOF=%s dt=%s EncryptKey=%s' % (self.blockNumber, self.currentBlockSize, self.stateEOF, str(time.time() - dt), self.keyID))
            return block

        d = maybeDeferred(_doBlock)
        d.addCallback(lambda block: self.automat('block-encrypted', block))
        d.addErrback(lambda err: self.automat('fail', err))

    def doBlockPushAndRaid(self, *args, **kwargs):
        """
        Action method.
        """
        newblock = args[0]
        if newblock is None:
            self.abort()
            self.automat('fail')
            if _Debug:
                lg.out(_DebugLevel, 'backup.doBlockPushAndRaid ERROR newblock is empty, terminating=%s' % self.terminating)
            lg.warn('failed to encrypt block, ABORTING')
            return
        if self.terminating:
            self.automat('block-raid-done', (newblock.BlockNumber, None))
            if _Debug:
                lg.out(_DebugLevel, 'backup.doBlockPushAndRaid SKIP, terminating=True')
            return
        fileno, filename = tmpfile.make('raid', extension='.raid')
        serializedblock = newblock.Serialize()
        blocklen = len(serializedblock)
        os.write(fileno, strng.to_bin(blocklen) + b':' + serializedblock)
        os.close(fileno)
        self.workBlocks[newblock.BlockNumber] = filename
        dt = time.time()
        outputpath = os.path.join(settings.getLocalBackupsDir(), self.customerGlobalID, self.pathID, self.version)
        task_params = (filename, self.eccmap.name, self.version, newblock.BlockNumber, outputpath)
        raid_worker.add_task('make', task_params, lambda cmd, params, result: self._raidmakeCallback(params, result, dt))
        self.automat('block-raid-started', newblock)
        del serializedblock
        if _Debug:
            lg.out(_DebugLevel, 'backup.doBlockPushAndRaid %s : start process data from %s to %s, %d' % (newblock.BlockNumber, filename, outputpath, id(self.terminating)))

    def doPopBlock(self, *args, **kwargs):
        """
        Action method.
        """
        blockNumber, _ = args[0]
        filename = self.workBlocks.pop(blockNumber)
        tmpfile.throw_out(filename, 'block raid done')

    def doFirstBlock(self, *args, **kwargs):
        """
        Action method.
        """
        self.dataSent = 0
        self.blocksSent = 0
        self.blockNumber = 0
        self.currentBlockSize = 0
        self.currentBlockData = BytesIO()

    def doNextBlock(self, *args, **kwargs):
        """
        Action method.
        """
        self.dataSent += self.currentBlockSize
        self.blocksSent += 1
        self.blockNumber += 1
        self.currentBlockSize = 0
        self.currentBlockData.close()
        self.currentBlockData = BytesIO()

    def doBlockReport(self, *args, **kwargs):
        """
        Action method.
        """
        BlockNumber, result = args[0]
        if self.blockResultCallback:
            self.blockResultCallback(self.backupID, BlockNumber, result)

    def doNotifyNewData(self, *args, **kwargs):
        """
        Action method.
        """
        if self.notifyNewDataCallback:
            self.notifyNewDataCallback()

    def doClose(self, *args, **kwargs):
        """
        Action method.
        """
        self.closed = True
        for filename in self.workBlocks.values():
            tmpfile.throw_out(filename, 'backup aborted')

    def doReport(self, *args, **kwargs):
        """
        Action method.
        """
        if self.ask4abort:
            if self.finishCallback:
                self.finishCallback(self.backupID, 'abort')
            self.resultDefer.callback('abort')
            events.send('backup-aborted', data=dict(backup_id=self.backupID, source_path=self.sourcePath))
        else:
            if self.finishCallback:
                self.finishCallback(self.backupID, 'done')
            self.resultDefer.callback('done')
            events.send('backup-done', data=dict(backup_id=self.backupID, source_path=self.sourcePath))

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        self._kill_pipe()
        self.eccmap = None
        self.pipe = None
        self.ask4abort = False
        self.terminating = False
        self.stateEOF = False
        self.stateReading = False
        self.closed = False
        self.workBlocks = None
        self.resultDefer = None
        self.finishCallback = None
        self.blockResultCallback = None
        self.currentBlockData.close()
        del self.currentBlockData
        self.destroy()
        collected = gc.collect()
        if _Debug:
            lg.out(_DebugLevel, 'backup.doDestroyMe [%s] collected %d objects' % (self.backupID, collected))

    def abort(self):
        """
        This method should stop this backup by killing the pipe process.
        """
        if _Debug:
            lg.out(_DebugLevel, 'backup.abort id %s, %d' % (str(self.backupID), id(self.ask4abort)))
        self.terminating = True
        for blockNumber, filename in self.workBlocks.items():
            lg.warn('aborting raid make worker for block %d in %s' % (blockNumber, filename))
            raid_worker.cancel_task('make', filename)
        lg.warn('killing backup pipe')
        self.ask4abort = True
        self._kill_pipe()

    def progress(self):
        """
        """
        if self.totalSize <= 0:
            return 0.0
        percent = min(100.0, 100.0*self.dataSent/self.totalSize)
        return percent

    def _raidmakeCallback(self, params, result, dt):
        _, _, _, blockNumber, _ = params
        if result is None:
            if _Debug:
                lg.out(_DebugLevel, 'backup._raidmakeCallback WARNING - result is None :  %r eof=%s dt=%s' % (blockNumber, str(self.stateEOF), str(time.time() - dt)))
            events.send('backup-raidmake-failed', data=dict(backup_id=self.backupID))
            self.ask4abort = True
            self._kill_pipe()
        else:
            if _Debug:
                lg.out(_DebugLevel, 'backup._raidmakeCallback %r %r eof=%s dt=%s' % (blockNumber, result, str(self.stateEOF), str(time.time() - dt)))
            self.automat('block-raid-done', (blockNumber, result))

    def _kill_pipe(self):
        if self.pipe:
            if _Debug:
                lg.out(_DebugLevel, 'backup._kill_pipe for %s' % self.backupID)
            try:
                self.pipe.kill()
            except:
                lg.exc()


#------------------------------------------------------------------------------


def main():
    from bitdust.system import bpio
    from bitdust.storage import backup_tar

    bpio.init()
    settings.init()

    lg.set_debug_level(24)
    lg.life_begins()

    automat.LifeBegins(lg.when_life_begins())
    automat.OpenLogFile(settings.AutomatsLog())

    key.InitMyKey()
    my_id.init()

    sourcePath = sys.argv[1]
    compress_mode = 'bz2'
    backupID = sys.argv[2]
    raid_worker.A('init')

    if bpio.pathIsDir(sourcePath):
        backupPipe = backup_tar.backuptardir_thread(sourcePath, compress=compress_mode)
    else:
        backupPipe = backup_tar.backuptarfile_thread(sourcePath, compress=compress_mode)

    def _bk_done(bid, result):
        from bitdust.crypt import signed
        customer, remotePath = packetid.SplitPacketID(bid)
        try:
            os.mkdir(os.path.join(settings.getLocalBackupsDir(), customer, remotePath + '.out'))
        except:
            pass
        for filename in os.listdir(os.path.join(settings.getLocalBackupsDir(), customer, remotePath)):
            filepath = os.path.join(settings.getLocalBackupsDir(), customer, remotePath, filename)
            payld = bpio.ReadBinaryFile(filepath)
            newpacket = signed.Packet('Data', my_id.getIDURL(), my_id.getIDURL(), filename, payld, 'http://megafaq.ru/cvps1010.xml')
            newfilepath = os.path.join(settings.getLocalBackupsDir(), customer, remotePath + '.out', filename)
            bpio.WriteBinaryFile(newfilepath, newpacket.Serialize())

    def _bk_closed(*args, **kwargs):
        # job.automat('fail')
        # del job
        reactor.stop()  # @UndefinedVariable

    def _bk_start():
        job = backup(backupID, backupPipe, blockSize=16*1024*1024)
        job.finishCallback = _bk_done  # lambda bid, result: _bk_done(bid, result, job)
        job.addStateChangedCallback(_bk_closed, oldstate=None, newstate='DONE')
        reactor.callLater(1, job.automat, 'start')  # @UndefinedVariable

    reactor.callLater(0, _bk_start)  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from CodernityDB.index import Index
import six
from six.moves import range
# from CodernityDB.env import cdb_environment
# import warnings


class ShardedIndex(Index):

    def __init__(self, db_path, name, *args, **kwargs):
        """
        There are 3 additional parameters. You have to hardcode them in your custom class. **NEVER** use directly

        :param int sh_nums: how many shards should be
        :param class ind_class: Index class to use (HashIndex or your custom one)
        :param bool use_make_keys: if True, `make_key`, and `make_key_value` will be overriden with those from first shard

        The rest parameters are passed straight to `ind_class` shards.

        """
        super(ShardedIndex, self).__init__(db_path, name)
        try:
            self.sh_nums = kwargs.pop('sh_nums')
        except KeyError:
            self.sh_nums = 5
        try:
            ind_class = kwargs.pop('ind_class')
        except KeyError:
            raise Exception("ind_class must be given")
        else:
            # if not isinstance(ind_class, basestring):
            #     ind_class = ind_class.__name__
            self.ind_class = ind_class
        if 'use_make_keys' in kwargs:
            self.use_make_keys = kwargs.pop('use_make_keys')
        else:
            self.use_make_keys = False
        self._set_shard_datas(*args, **kwargs)
        self.patchers = []  # database object patchers

    def _set_shard_datas(self, *args, **kwargs):
        self.shards = {}
        self.shards_r = {}
#        ind_class = globals()[self.ind_class]
        ind_class = self.ind_class
        i = 0
        for sh_name in [self.name + str(x) for x in range(self.sh_nums)]:
            # dict is better than list in that case
            self.shards[i] = ind_class(self.db_path, sh_name, *args, **kwargs)
            self.shards_r['%02x' % i] = self.shards[i]
            self.shards_r[i] = self.shards[i]
            i += 1

        if not self.use_make_keys:
            self.make_key = self.shards[0].make_key
            self.make_key_value = self.shards[0].make_key_value

        self.last_used = 0

    @property
    def storage(self):
        st = self.shards[self.last_used].storage
        return st

    def __getattr__(self, name):
        return getattr(self.shards[self.last_used], name)

    def open_index(self):
        for curr in six.itervalues(self.shards):
            curr.open_index()

    def create_index(self):
        for curr in six.itervalues(self.shards):
            curr.create_index()

    def destroy(self):
        for curr in six.itervalues(self.shards):
            curr.destroy()

    def compact(self):
        for curr in six.itervalues(self.shards):
            curr.compact()

    def reindex(self):
        for curr in six.itervalues(self.shards):
            curr.reindex()

    def all(self, *args, **kwargs):
        for curr in six.itervalues(self.shards):
            for now in curr.all(*args, **kwargs):
                yield now

    def get_many(self, *args, **kwargs):
        for curr in six.itervalues(self.shards):
            for now in curr.get_many(*args, **kwargs):
                yield now



#!/usr/bin/env python
# strng.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (strng.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: strng.

"""

#------------------------------------------------------------------------------

import sys

#------------------------------------------------------------------------------

if sys.version_info[0] == 3:
    text_type = str
    binary_type = bytes
else:
    text_type = unicode  # @UndefinedVariable
    binary_type = str

#------------------------------------------------------------------------------


def is_text(s):
    """
    Return `True` if `s` is a text value:
    + `unicode()` in Python2
    + `str()` in Python3
    """
    return isinstance(s, text_type)


def is_bin(s):
    """
    Return `True` if `s` is a binary value:
    + `str()` in Python2
    + `bytes()` in Python3
    """
    return isinstance(s, binary_type)


def is_string(s):
    """
    Return `True` if `s` is text or binary type (not integer, class, list, etc...)
    """
    return is_text(s) or is_bin(s)


#------------------------------------------------------------------------------


def to_text(s, encoding='utf-8', errors='strict'):
    """
    If ``s`` is binary type - decode it to unicode - "text" type in Python3 terms.
    If ``s`` is not binary and not text calls `str(s)` to build text representation.
    """
    if s is None:
        return s
    if not is_string(s):
        s = '{0}'.format(s)
    if is_text(s):
        return s
    return s.decode(encoding=encoding, errors=errors)


def to_bin(s, encoding='utf-8', errors='strict'):
    """
    If ``s`` is unicode ("text" type in Python3 terms) - encode it to utf-8, otherwise return ``s``.
    If ``s`` is not binary and not text calls `str(s)` to build text representation,
    then encode to binary and return.
    """
    if s is None:
        return s
    if not is_string(s):
        s = '{0}'.format(s)
    if is_bin(s):
        return s
    return s.encode(encoding=encoding, errors=errors)



#!/usr/bin/python
# encrypted.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (encrypted.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: encrypted.

Higher level code interfaces with ``encrypted`` so that it does not have to deal
with ECC stuff.  We write or read a large block at a time (maybe 64 MB say).
When writing we generate all the ECC information, and when reading we will
use ECC to recover lost information so user sees whole block still.

We have to go to disk. The normal mode is probably that there are a few machines that
are slow and the rest move along. We want to get the backup secure as soon as possible.
It can be secure even if 5 to 10 suppliers are not finished yet.  But this could be
a lot of storage, so we should be using disk.

We want to generate a pool of writes to do, and put in more as it gets below some
MB limit. But we should not be limited by a particular nodes speed.

The ``packet`` will have all the info about where it is going etc.
We number them with our block number and the supplier numbers.

Going to disk should let us do restarts after crashes without much trouble.

Digital signatures and timestamps are done on ``encrypted`` blocks of data.
Signatures are also done on ``packets`` and ``encrypted`` blocks.

RAIDMAKE:
    This object can be asked to generate any/all ``packet(s)`` that would come from this ``encrypted``.
RAIDREAD:
    It can also rebuild the ``encrypted`` from packets and will
    generate the read requests to get fetch the packets.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import traceback
import base64

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import serialization

from bitdust.userid import my_id
from bitdust.userid import id_url

from bitdust.crypt import key
from bitdust.crypt import my_keys

#------------------------------------------------------------------------------


class Block(object):

    """
    A class to represent an encrypted Data block. The only 2 things secret in
    here will be the ``EncryptedSessionKey`` and ``EncryptedData``. Scrubbers
    may combine-packets/unserialize/inspect-blocks/check-signatures.

    CreatorID              http://cate.com/id1.xml  - so people can check signature - says PK type too
    BackupID               Creator's ID for the backup this packet is part of
    BlockNumber            number of this block
    EncryptedData          data may have some padding so multiple of crypto chunck size
                           and multiple of #nodes in eccmap (usually 64) for division
                           into packets
    Length                 real length of data when cleartext (encrypted may be padded)
    LastBlock              should now be "True" or "False" - careful in using
    SessionKeyType         which crypto is used for session key
    EncryptedSessionKey    encrypted with our public key so only we can read this
    Other                  could be be for professional timestamp company or other future features
    Signature              digital signature by Creator - verifiable by public key in creator identity
    """

    def __init__(
        self,
        CreatorID=None,
        BackupID='',
        BlockNumber=0,
        SessionKey='',
        SessionKeyType=None,
        LastBlock=True,
        Data=b'',
        EncryptKey=None,
        DecryptKey=None,
        EncryptedSessionKey=None,
        EncryptedData=None,
        Length=None,
        Signature=None,
    ):
        self.CreatorID = CreatorID
        if not self.CreatorID:
            self.CreatorID = my_id.getIDURL()
        if not isinstance(self.CreatorID, id_url.ID_URL_FIELD):
            self.CreatorID = id_url.field(self.CreatorID)
        self.BackupID = strng.to_text(BackupID)
        self.BlockNumber = BlockNumber
        self.LastBlock = bool(LastBlock)
        self.SessionKeyType = SessionKeyType or key.SessionKeyType()
        if EncryptedSessionKey:
            # this block to be decrypted after receiving
            self.EncryptedSessionKey = EncryptedSessionKey
        else:
            # this block to be encrypted before sending
            if callable(EncryptKey):
                self.EncryptedSessionKey = EncryptKey(SessionKey)
            elif strng.is_text(EncryptKey):
                self.EncryptedSessionKey = my_keys.encrypt(EncryptKey, SessionKey)
            elif strng.is_bin(EncryptKey):
                self.EncryptedSessionKey = my_keys.encrypt(strng.to_text(EncryptKey), SessionKey)
            else:
                self.EncryptedSessionKey = key.EncryptLocalPublicKey(SessionKey)
        if EncryptedData and Length is not None:
            self.Length = Length
            self.EncryptedData = EncryptedData
        else:
            self.Length = len(Data)
            self.EncryptedData = key.EncryptWithSessionKey(SessionKey, Data, session_key_type=self.SessionKeyType)
        if Signature:
            self.Signature = Signature
        else:
            self.Signature = None
            self.Sign(signing_key=EncryptKey)
        self.DecryptKey = DecryptKey
        if _Debug:
            lg.out(_DebugLevel, 'new data in %s' % self)

    def __repr__(self):
        return 'encrypted{ BackupID=%s BlockNumber=%s Length=%s LastBlock=%s }' % (str(self.BackupID), str(self.BlockNumber), str(self.Length), self.LastBlock)

    def SessionKey(self):
        """
        Return original SessionKey from ``EncryptedSessionKey`` using one of the methods
        depend on the type of ``DecryptKey`` parameter passed in the __init__()

            + ``crypt.key.DecryptLocalPrivateKey()`` if DecryptKey is None
            + ``my_keys.decrypt()`` if DecryptKey is a string with key_id
            + ``DecryptKey()`` if this is a callback method
        """
        if callable(self.DecryptKey):
            return self.DecryptKey(self.EncryptedSessionKey)
        elif strng.is_text(self.DecryptKey):
            return my_keys.decrypt(self.DecryptKey, self.EncryptedSessionKey)
        elif strng.is_bin(self.DecryptKey):
            return my_keys.decrypt(strng.to_text(self.DecryptKey), self.EncryptedSessionKey)
        return key.DecryptLocalPrivateKey(self.EncryptedSessionKey)

    def GenerateHashBase(self):
        """
        Generate a single string with all data fields, used to create a hash
        for that ``encrypted_block``.
        """
        sep = b'::::'
        StringToHash = b''
        StringToHash += self.CreatorID.to_original()
        StringToHash += sep + strng.to_bin(self.BackupID)
        StringToHash += sep + strng.to_bin(str(self.BlockNumber))
        StringToHash += sep + strng.to_bin(self.SessionKeyType)
        StringToHash += sep + strng.to_bin(self.EncryptedSessionKey)
        StringToHash += sep + strng.to_bin(str(self.Length))
        StringToHash += sep + strng.to_bin(str(self.LastBlock))
        StringToHash += sep + strng.to_bin(self.EncryptedData)
        return StringToHash

    def GenerateHash(self):
        """
        Create a hash for that ``encrypted_block`` using ``crypt.key.Hash()``.
        """
        return key.Hash(self.GenerateHashBase())

    def Sign(self, signing_key):
        """
        Generate digital signature for that ``encrypted_block``.
        """
        # usually just done at packet creation
        self.Signature = self.GenerateSignature(signing_key)
        return self

    def GenerateSignature(self, signing_key):
        """
        Call ``crypt.key.Sign()`` to generate signature.
        """
        # TODO: make possible to use `signing_key` to sign data
        return key.Sign(self.GenerateHash())

    def Ready(self):
        """
        Just return True if signature is already created.
        """
        return self.Signature is not None

    def Valid(self):
        """
        Validate signature to verify the ``encrypted_block``.
        Not used at the moment.
        """
        if not self.Ready():
            lg.warn('block is not ready yet ' + str(self))
            return False
        # TODO: make possible to verify signature using `signing_key`
        hashsrc = self.GenerateHash()
        ConIdentity = my_id.getLocalIdentity()
        if ConIdentity is None:
            lg.warn('could not get Identity so returning False')
            return False
        result = key.Verify(ConIdentity, hashsrc, self.Signature)  # At block level only work on own stuff
        return result

    def Data(self):
        """
        Return an original data, decrypt using ``EncryptedData`` and
        ``EncryptedSessionKey``.
        """
        SessionKey = self.SessionKey()
        ClearLongData = key.DecryptWithSessionKey(SessionKey, self.EncryptedData, session_key_type=self.SessionKeyType)
        return ClearLongData[0:self.Length]  # remove padding

    def Serialize(self):
        """
        Create a string that stores all data fields of that ``encrypted.Block``
        object.
        """
        dct = {
            'c': self.CreatorID.to_text(),
            'b': self.BackupID,
            'n': self.BlockNumber,
            'e': self.LastBlock,
            'k': strng.to_text(base64.b64encode(strng.to_bin(self.EncryptedSessionKey))),
            't': self.SessionKeyType,
            'l': self.Length,
            'p': self.EncryptedData,
            's': self.Signature,
        }
        if _Debug:
            lg.out(_DebugLevel, 'encrypted.Serialize %s' % repr(dct)[:100])
        return serialization.DictToBytes(dct, encoding='utf-8')


#------------------------------------------------------------------------------


def Unserialize(data, decrypt_key=None):
    """
    A method to create a ``encrypted.Block`` instance from input string.
    """
    dct = serialization.BytesToDict(data, keys_to_text=True, encoding='utf-8')
    if _Debug:
        lg.out(_DebugLevel, 'encrypted.Unserialize %s' % repr(dct)[:100])
    try:
        _c = dct['c']
        _p = dct['p']
        _n = dct['n']
        _e = dct['e']
        _k = dct['k']
        _t = dct['t']
        _l = dct['l']
        _s = dct['s']
        _b = dct['b']
    except Exception as exc:
        lg.exc('data unserialize failed with %r: %r\n%r\n%s' % (
            exc,
            list(dct.keys()),
            (dct.get('c'), dct.get('b'), dct.get('i'), dct.get('r')),
            traceback.format_tb(),
        ))
        if _Debug:
            lg.out(_DebugLevel, repr(dct))
        return None
    try:
        newobject = Block(
            CreatorID=id_url.field(_c),
            BackupID=strng.to_text(_b),
            BlockNumber=_n,
            LastBlock=_e,
            EncryptedSessionKey=base64.b64decode(strng.to_bin(_k)),
            SessionKeyType=strng.to_text(_t),
            Length=_l,
            EncryptedData=_p,
            Signature=_s,
            DecryptKey=decrypt_key,
        )
    except:
        lg.exc()
        return None
    return newobject
